# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #
from typing import Literal


def mpl_display_image(image_data):
    import io
    import matplotlib.pyplot as plt
    import matplotlib.image as mpimg

    sio = io.BytesIO()
    sio.write(image_data)
    sio.seek(0)
    img = mpimg.imread(sio)
    plt.imshow(img, aspect="auto")
    plt.xticks([])
    plt.yticks([])


def os_display_image(image_data, ftype="png", remove_file=True):
    import os
    from sys import platform

    fname = f"/tmp/out.{ftype}"
    with open(fname, "wb") as fh:
        fh.write(image_data)

    if remove_file:
        remove_cmd = f"rm {fname}"
    else:
        remove_cmd = "true"

    if platform == "darwin":
        # open doesn't block, wait a little not to remove the file too early
        os.system(f"(open {fname} && sleep 5 && {remove_cmd})&")
    else:
        os.system(f"(display {fname} && {remove_cmd})&")


def ipy_display_image(image_data):
    from IPython.display import Image, display

    plt = Image(image_data)
    display(plt)


def plot_pdot(
        pdot,
        backend: Literal['mpl', 'os', 'ipy'] = 'ipy',
        ftype: Literal['png', 'svg', 'pdf'] = 'png',
        remove_file: bool = True,
):
    if ftype == "png":
        image_data = pdot.create_png()
    elif ftype == "svg":
        image_data = pdot.create_svg()
    elif ftype == "pdf":
        image_data = pdot.create_pdf()
    else:
        raise ValueError(f"Unknown ftype: {ftype}")

    if backend == "mpl" and ftype == "png":
        mpl_display_image(image_data)
    elif backend == "os":
        os_display_image(image_data, ftype=ftype, remove_file=remove_file)
    elif backend == "ipy":
        ipy_display_image(image_data)
    else:
        raise ValueError(f"Unknown backend: {backend} or combination of "
                         f"ftype ({ftype}) and backend invalid.")

    return pdot


class AttrDict(dict):
    def __init__(self, *args, **kwargs):
        super(AttrDict, self).__init__(*args, **kwargs)
        self.__dict__ = self



# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2023.                 #
# ######################################### #
import numpy as np
import pickle
import pytest

import xdeps as xd
import xdeps.tasks


def test_manager_refs():
    manager = xd.Manager()
    container1 = {'b': 3}
    container2 = {}
    ref1 = manager.ref(container1, 'ref1')
    ref2 = manager.ref(container2, 'ref2')

    ref1['a'] = 1
    ref2['a'] = 2

    assert container1['a'] == 1
    assert ref1['a']._get_value() == 1
    assert container2['a'] == 2
    assert ref2['a']._get_value() == 2

    assert container1['b'] == 3
    assert ref1['b']._get_value() == 3


def test_expr_task():
    manager = xd.Manager()
    container = {'b': 3, 'c': 4}
    ref = manager.ref(container, 'ref')
    target = ref['a']
    expr = 2 * ref['b'] + 3 * ref['c']

    expr_task = xdeps.tasks.ExprTask(target, expr)

    assert expr_task.expr == expr
    assert expr_task.taskid == target
    assert expr_task.dependencies == {ref['b'], ref['c']}
    assert expr_task.targets == {target}

    assert container == {'b': 3, 'c': 4}
    expr_task.run()
    assert container == {'a': 18, 'b': 3, 'c': 4}


def test_depenv():
    manager = xd.Manager()
    container = {'b': 3, 'c': 4}
    ref = manager.ref(container, 'ref')
    ref['a'] = ref['b'] + ref['c']

    depenv = xd.tasks.DepEnv(container, ref)

    assert (depenv['b'] == 3) is True
    assert (depenv['c'] == 4) is True
    assert (depenv['a'] == 7) is True
    assert set(depenv.keys()) == {'a', 'b', 'c'}

    depenv['b'] = 5
    depenv.c = 6
    assert (depenv['a'] == 11) is True

    assert (depenv._eval('a + b')._value == 16) is True


@pytest.fixture(scope='function')
def example_manager():
    manager = xd.Manager()

    container1 = {'a': 12, 'b': 8, 'c': -4, 'g': 42}
    ref1 = manager.ref(container1, 'ref1')

    container2 = {'d': [2, 3]}
    ref2 = manager.ref(container2, 'ref2')

    # We are realising the following setup in the test:
    #
    #            ╭ [a]─┐                                      Legend:
    #            │     │                                      [x] ref/expr x
    # container1 │ [b]─┴─►(f=a+b)─►[f]─┐                      (x) task x
    #            │                     │
    #            ╰ [c]───►(g=-c)─►[g]──┴─►(h=√(f-g))─►[h]─┐
    #                                                     │
    #                                                     │
    #            ╭ [d[0]]─┐                               │
    # container2 │        │                               │
    #            ╰ [d[1]]─┴───►(e[0]=d[0]*d[1])───►[e[0]]─┴─►(j=h+e[0])─►[j]

    ref1['f'] = ref1['a'] + ref1['b']

    ref1['h'] = (ref1['f'] - ref1['g']) ** 0.5

    # Define [g] after use to test to fully test manager.register with rtasks.
    ref1['g'] = -ref1['c']

    ref2['e'] = [None]
    ref2['e'][0] = ref2['d'][0] * ref2['d'][1]

    ref2['j'] = ref1['h'] + ref2['e'][0]

    return manager, [ref1, ref2], [container1, container2]


def test_manager(example_manager):
    manager, refs, containers = example_manager
    ref1, ref2 = refs
    container1, container2 = containers

    # First we check that everything was registered correctly
    f_expr = ref1['a'] + ref1['b']
    h_expr = (ref1['f'] - ref1['g']) ** 0.5
    g_expr = -ref1['c']
    e0_expr = ref2['d'][0] * ref2['d'][1]
    j_expr = ref1['h'] + ref2['e'][0]

    # Check that tasks are correctly created.
    assert len(manager.tasks) == 5
    assert manager.tasks[ref1['f']].expr.equals(f_expr)
    assert manager.tasks[ref1['g']].expr.equals(g_expr)
    assert manager.tasks[ref1['h']].expr.equals(h_expr)
    assert manager.tasks[ref2['e'][0]].expr.equals(e0_expr)
    assert manager.tasks[ref2['j']].expr.equals(j_expr)

    # Check that rdeps are correctly created.
    assert len(manager.rdeps) == 11
    assert set(manager.rdeps[ref1['a']]) == {ref1['f']}  # [f] depends on [a]
    assert set(manager.rdeps[ref1['b']]) == {ref1['f']}  # [f] depends on [b]
    assert set(manager.rdeps[ref1['c']]) == {ref1['g']}  # [g] depends on [c]
    assert set(manager.rdeps[ref1['f']]) == {ref1['h']}  # [h] depends on [f]
    assert set(manager.rdeps[ref1['g']]) == {ref1['h']}  # [h] depends on [g]
    # [e[0]] and thus [e] depend on [d[0]], [d[1]], and thus [d]:
    assert set(manager.rdeps[ref2['d']]) == {ref2['e'], ref2['e'][0]}
    assert set(manager.rdeps[ref2['d'][0]]) == {ref2['e'], ref2['e'][0]}
    assert set(manager.rdeps[ref2['d'][1]]) == {ref2['e'], ref2['e'][0]}
    assert set(manager.rdeps[ref1['h']]) == {ref2['j']}  # [j] depends on [h]
    assert set(manager.rdeps[ref2['e'][0]]) == {ref2['j']}  # [j] depends on [e[0]]
    assert set(manager.rdeps[ref2['e']]) == {ref2['j']}  # thus [j] depends on [e]

    # Check that rtasks are correctly created.
    # This is a special case as all tasks so far output one ref.
    # In this case is subset of rdeps as it has taskids on the lhs (and since
    # not all refs are tasks, we have fewer entries than in rdeps).
    assert len(manager.rtasks) == 4
    assert set(manager.rtasks[ref1['h']]) == {ref2['j']}
    assert set(manager.rtasks[ref2['e'][0]]) == {ref2['j']}
    assert set(manager.rtasks[ref1['f']]) == {ref1['h']}
    assert set(manager.rtasks[ref1['g']]) == {ref1['h']}

    # Check that deptasks are correctly created.
    # Almost the same as rdeps but maps to taskids instead of refs, so:
    # - We have an entry for [j],
    # - [e[0]] etc. are on the rhs but not [e].
    assert len(manager.deptasks) == 12
    assert set(manager.deptasks[ref1['a']]) == {ref1['f']}
    assert set(manager.deptasks[ref1['b']]) == {ref1['f']}
    assert set(manager.deptasks[ref1['c']]) == {ref1['g']}
    assert set(manager.deptasks[ref1['f']]) == {ref1['h']}
    assert set(manager.deptasks[ref1['g']]) == {ref1['h']}
    assert set(manager.deptasks[ref2['d']]) == {ref2['e'][0]}
    assert set(manager.deptasks[ref2['d'][0]]) == {ref2['e'][0]}
    assert set(manager.deptasks[ref2['d'][1]]) == {ref2['e'][0]}
    assert set(manager.deptasks[ref1['h']]) == {ref2['j']}
    assert set(manager.deptasks[ref2['e'][0]]) == {ref2['j']}
    assert set(manager.deptasks[ref2['e']]) == {ref2['j']}
    assert set(manager.deptasks[ref2['j']]) == set()

    # Check that tartasks are correctly created.
    # In this case it is (roughly) an identity on refs that are tasks,
    # and zero otherwise.
    assert len(manager.tartasks) == 12
    # Tasks:
    for ref in [ref1['f'], ref1['g'], ref1['h'], ref2['e'][0], ref2['j']]:
        assert set(manager.tartasks[ref]) == {ref}
    # Extra considerations for e for e[0] etc.:
    assert set(manager.tartasks[ref2['e']]) == {ref2['e'][0]}
    # Refs that aren't tasks:
    for ref in [ref1['a'], ref1['b'], ref1['c'], ref2['d'], ref2['d'][0], ref2['d'][1]]:
        assert set(manager.tartasks[ref]) == set()

    # Check that containers are correctly created.
    assert manager.containers == {'ref1': ref1, 'ref2': ref2}

    # Now we check that the manager can correctly execute the tasks
    assert container1['f'] == 20
    assert container1['g'] == 4
    assert container1['h'] == 4
    assert container2['e'][0] == 6
    assert container2['j'] == 10

    # Test that values are updated correctly
    ref1['a'] = 21
    assert container2['j'] == 11


def test_manager_unregister(example_manager):
    manager, refs, containers = example_manager
    ref1, ref2 = refs
    container1, container2 = containers

    # Test unregistering a task
    ref1['h'] = 4
    assert container2['j'] == 10


def test_manager_clone_verify_refresh(example_manager):
    manager, refs, _ = example_manager
    ref1, ref2 = refs

    # Test consistency and make a copy
    manager.verify()
    manager_copy = manager.copy()

    # Mess up consistency and check again
    manager.rtasks[ref1['f']].append(ref2['e'][0])
    manager.rdeps[ref1['b']].remove(ref1['f'])

    with pytest.raises(ValueError) as error:
        manager.verify()
    assert 'not consistent' in str(error.value)

    manager_copy.verify()  # the copy should still be okay

    # Refresh and check again
    manager.refresh()
    manager.verify()


def test_manager_dump_and_load(example_manager):
    manager, _, _ = example_manager

    dumped = manager.dump()

    new_manager = xd.Manager()
    new_container1 = {'a': 10, 'b': 10, 'c': -4}
    new_ref1 = new_manager.ref(new_container1, 'new_ref1')
    new_container2 = {'d': [3, 2], 'e': [-99]}
    new_ref2 = new_manager.ref(new_container2, 'new_ref2')

    old_expr = new_ref2['j'] = 2 * new_ref1['a']  # this will be removed by load

    new_containers = {
        'ref1': new_ref1,
        'ref2': new_ref2,
    }
    new_manager.load(dumped, new_containers)

    assert new_containers['ref2']['j']._get_value() == 20
    assert not new_manager.tasks[new_ref2['j']].expr == old_expr


def test_manager_pickle(example_manager):
    manager, _, original_containers = example_manager
    original1, original2 = original_containers

    pickled = pickle.dumps(manager)
    new_manager = pickle.loads(pickled)

    ref1 = new_manager.containers['ref1']
    container1 = ref1._owner
    ref2 = new_manager.containers['ref2']
    container2 = ref2._owner

    ref1['a'] = 32
    ref2['d'][1] = 5

    assert container1 is not original1
    assert container2 is not original2
    assert container1['f'] == 40
    assert container1['h'] == 6
    assert container2['e'][0] == 10
    assert container2['j'] == 16


def test_ref_count():
    manager = xd.Manager()
    container = {'a': 3, 'b': [None], 'c': None}
    ref = manager.ref(container, 'ref')

    ref['c'] = ref['b'][0]
    ref['b'][0] = ref['a']

    assert len(manager.rtasks) == 1
    assert len(manager.rtasks[ref['b'][0]]) == 1
    assert manager.rtasks[ref['b'][0]][ref['c']] == 2
    # This is a bit weird, the task (b[0]) influences [c] twice.
    # One time directly, and the other comes from the fact that the task
    # (b[0]) influences both the ref [b[0]] and [b]. Therefore both of these
    # refs, in turn, influence [c].
    # The setup of this example requires [a], as we need two tasks, the first
    # one being (b[0]) and the second one being (a).

    assert container == {'a': 3, 'b': [3], 'c': 3}


def test_function_task():
    manager = xd.Manager()
    container = {'a': 3, 'b': 4}
    ref = manager.ref(container, 'ref')

    def action():
        ref['c'] = ref['a']._get_value() + ref['b']._get_value()
        ref['d'] = ref['a']._get_value() * ref['b']._get_value()

    task = xd.tasks.FunctionTask(
        taskid='my_action',
        action=action,
        targets={ref['c'], ref['d']},
        dependencies={ref['a'], ref['b']},
    )

    manager.register(task)
    task.run()

    assert container == {'a': 3, 'b': 4, 'c': 7, 'd': 12}

    ref['a'] = 4
    assert container == {'a': 4, 'b': 4, 'c': 8, 'd': 16}

    assert len(manager.tasks) == 1
    assert manager.tasks['my_action'] is task

    assert len(manager.rdeps) == 2
    assert set(manager.rdeps[ref['a']]) == {ref['c'], ref['d']}
    assert set(manager.rdeps[ref['b']]) == {ref['c'], ref['d']}

    assert len(manager.deptasks) == 4
    assert set(manager.deptasks[ref['a']]) == {'my_action'}
    assert set(manager.deptasks[ref['b']]) == {'my_action'}
    assert set(manager.deptasks[ref['c']]) == set()
    assert set(manager.deptasks[ref['d']]) == set()

    assert len(manager.tartasks) == 4
    assert set(manager.tartasks[ref['a']]) == set()
    assert set(manager.tartasks[ref['b']]) == set()
    assert set(manager.tartasks[ref['c']]) == {'my_action'}
    assert set(manager.tartasks[ref['d']]) == {'my_action'}


@pytest.mark.xfail(reason='Not implemented yet, unclear if it should be')
def test_function_task_depends_on_container():
    manager = xd.Manager()
    container_source = {'a': 3, 'b': 4}
    container_target = {}
    ref_source = manager.ref(container_source, 'ref_source')
    ref_target = manager.ref(container_target, 'ref_target')

    def action():
        ref_target['c'] = ref_source['a']._get_value() + ref_source['b']._get_value()
        ref_target['d'] = ref_source['a']._get_value() * ref_source['b']._get_value()

    task = xd.tasks.FunctionTask(
        taskid='my_action',
        action=action,
        targets={ref_target['c'], ref_target['d']},
        dependencies={ref_source},
    )

    manager.register(task)
    task.run()

    assert container_target == {'c': 7, 'd': 12}
    ref_source['a'] = 4
    assert container_target == {'c': 8, 'd': 16}


def test_lsa_knob_example():
    manager = xd.Manager()
    container = {'src': 10, 'tar': list(range(5))}
    ref = manager.ref(container, 'ref')

    weights = np.linspace(0, 1, 5)
    lsa = xd.tasks.LinearKnob(
        taskid='lsa',
        source=ref['src'],
        weights=weights,
        targets=[ref['tar'][ii] for ii in range(5)],
    )

    manager.register(lsa)
    lsa.run()

    assert container['tar'] == [0, 1, 2, 3, 4]

    ref['src'] = 20
    assert container['tar'] == [0.0, 3.5, 7.0, 10.5, 14.0]



import copy
import re
import logging

import numpy as np
from ..general import _print

#from scipy.optimize import fsolve, minimize
from .jacobian import JacobianSolver
from ..table import Table

log = logging.getLogger(__name__)

LIMITS_DEFAULT = (-1e200, 1e200)
STEP_DEFAULT = 1e-10
TOL_DEFAULT = 1e-10


class Vary:
    def __init__(
        self,
        name,
        container,
        limits=None,
        step=None,
        weight=None,
        max_step=None,
        tag="",
        active=True,
    ):

        if weight is None:
            weight = 1.0

        if limits is not None:
            assert len(limits) == 2, "`limits` must have length 2."
            limits = np.array(limits)

        assert weight > 0, "`weight` must be positive."

        self.name = name
        self.limits = limits
        self.step = step
        self.weight = weight
        self.container = container
        self.max_step = max_step
        self.active = active
        self.tag = tag

        self._complete_limits_and_step_from_defaults()

    def get_value(self):
        val = self.container[self.name]
        if hasattr(val, "_value"):
            return val._value
        else:
            return val

    def _complete_limits_and_step_from_defaults(self):
        if (
            self.limits is None
            and hasattr(self.container, "vary_default")
            and self.name in self.container.vary_default
        ):
            self.limits = self.container.vary_default[self.name]["limits"]

        if (
            self.step is None
            and hasattr(self.container, "vary_default")
            and self.name in self.container.vary_default
        ):
            self.step = self.container.vary_default[self.name]["step"]

    def __repr__(self):
        try:
            lim = f"({self.limits[0]:.4g}, {self.limits[1]:.4g})"
        except (IndexError, TypeError):
            lim = self.limits
        try:
            step = f"{self.step:.4g}"
        except (ValueError, TypeError):
            step = self.step
        try:
            weight = f"{self.weight:.4g}"
        except (ValueError, TypeError):
            weight = self.weight
        return f"Vary(name={self.name!r}, limits={lim}, step={step}, weight={weight})"


class VaryList:
    def __init__(self, vars, container, **kwargs):
        self.vary_objects = [Vary(vv, container, **kwargs) for vv in vars]


class Target:
    def __init__(
        self,
        tar,
        value,
        tol=None,
        weight=None,
        scale=None,
        action=None,
        tag="",
        optimize_log=False,
    ):

        if scale is not None and weight is not None:
            raise ValueError("Cannot specify both `weight` and `scale` for a target.")

        if scale is not None:
            weight = scale

        self.tar = tar
        self.action = action
        self.value = value
        self.tol = tol
        self.weight = weight
        self.active = True
        self.tag = tag
        self.optimize_log = optimize_log

    def __repr__(self):
        out = "Target("
        if callable(self.tar):
            tar_repr = "callable"
        else:
            tar_repr = repr(self.tar)
        try:
            val_str = f"{self.value:.6g}"
        except (ValueError, TypeError):
            val_str = self.value
        try:
            tol_str = f"{self.tol:.4g}"
        except (ValueError, TypeError):
            tol_str = self.tol
        try:
            weight_str = f"{self.weight:.4g}"
        except (ValueError, TypeError):
            weight_str = self.weight
        out += f"{tar_repr}, val={val_str}, tol={tol_str}, weight={weight_str}"
        if self.optimize_log:
            out += ", optimize_log=True"
        out += ")"
        return out

    def copy(self):
        return copy.copy(self)

    @property
    def scale(self):
        return self.weight

    @scale.setter
    def scale(self, value):
        self.weight = value

    def eval(self, data):
        res = data[self.action]
        if callable(self.tar):
            return self.tar(res)
        else:
            return res[self.tar]

    def runeval(self):
        return self.eval({self.action: self.action.run()})


class TargetList:
    def __init__(self, tars, **kwargs):
        self.targets = [Target(tt, **kwargs) for tt in tars]


class Action:
    def prepare(self):
        pass

    def run(self):
        return dict()

    def target(self, tar, value, **kwargs):
        return Target(tar, value, action=self, **kwargs)


class MeritFunctionForMatch:

    def __init__(
        self,
        vary,
        targets,
        actions,
        return_scalar,
        call_counter,
        verbose,
        tw_kwargs,
        steps_for_jacobian,
        check_limits,
        show_call_counter=True,
    ):

        self.vary = vary
        self.targets = targets
        self.actions = actions
        self.return_scalar = return_scalar
        self.call_counter = call_counter
        self.verbose = verbose
        self.tw_kwargs = tw_kwargs
        self.steps_for_jacobian = steps_for_jacobian
        self.found_point_within_tol = False
        self.zero_if_met = False
        self.show_call_counter = show_call_counter
        self.check_limits = check_limits

    def _x_to_knobs(self, x):
        knob_values = np.array(x).copy()
        for ii, vv in enumerate(self.vary):
            if vv.weight is not None:
                knob_values[ii] *= vv.weight
        return knob_values

    def _knobs_to_x(self, knob_values):
        x = np.array(knob_values, dtype=np.float64).copy()
        for ii, vv in enumerate(self.vary):
            if vv.weight is not None:
                x[ii] /= vv.weight
        return x

    def _extract_knob_values(self):
        return [ vv.get_value() for vv in self.vary]

    def _get_x(self):
        return self._knobs_to_x(self._extract_knob_values())

    def _set_x(self, x):
        self(x)  # this sets the knobs

    def _get_x_limits(self):
        knob_limits = []
        for vv in self.vary:
            if vv.limits is None:
                knob_limits.append(LIMITS_DEFAULT)
            else:
                knob_limits.append(vv.limits)
        knob_limits = np.array(knob_limits)
        x_lim_low = self._knobs_to_x(np.atleast_1d(np.squeeze(knob_limits[:, 0])))
        x_lim_high = self._knobs_to_x(np.atleast_1d(np.squeeze(knob_limits[:, 1])))
        x_limits = np.array([[hh, ll] for hh, ll in zip(x_lim_low, x_lim_high)])
        return x_limits

    def get_merit_function(self, check_limits=True, return_scalar=None):
        return MeritFuctionView(
            self, check_limits=check_limits, return_scalar=return_scalar
        )

    def __call__(self, x=None, check_limits=None, return_scalar=None):

        if x is None:
            knob_values = self._extract_knob_values()
        else:
            knob_values = self._x_to_knobs(x)

        if check_limits is None:
            check_limits = self.check_limits

        # Set knobs
        for vv, val in zip(self.vary, knob_values):
            if vv.active:
                if check_limits and vv.limits is not None and vv.limits[0] is not None:
                    if val < vv.limits[0]:
                        raise ValueError(f"Knob {vv.name} is below lower limit.")
                if check_limits and vv.limits is not None and vv.limits[1] is not None:
                    if val > vv.limits[1]:
                        raise ValueError(f"Knob {vv.name} is above upper limit.")
                vv.container[vv.name] = val

        # if self.verbose:
        #     _print(f'x = {knob_values}')

        # Run actions
        res_data = {}
        failed = False
        for aa in self.actions:
            res_data[aa] = aa.run()
            if isinstance(res_data[aa], str) and res_data[aa] == "failed":
                failed = True
                break

        if failed:
            err_values = [1e100 for tt in self.targets]
        else:
            res_values = []
            target_values = []
            for tt in self.targets:
                res_values.append(tt.eval(res_data))
                if hasattr(tt.value, "_value"):
                    target_values.append(tt.value._value)
                else:
                    target_values.append(tt.value)
            self._last_data = res_data  # for debugging

            res_values = np.array(res_values)
            target_values = np.array(target_values)

            transformed_res_values = res_values * 0
            for ii, tt in enumerate(self.targets):
                if hasattr(tt, "transform"):
                    transformed_res_values[ii] = tt.transform(res_values[ii])
                else:
                    transformed_res_values[ii] = res_values[ii]

            err_values = transformed_res_values - target_values

            # if self.verbose:
            #     _print(f'   f(x) = {res_values}')

            tols = 0 * err_values
            for ii, tt in enumerate(self.targets):
                tols[ii] = tt.tol

            # if self.verbose:
            #     _print(f'   err/tols = {err_values/tols}')

            targets_within_tol = np.abs(err_values) < tols
            self.last_targets_within_tol = targets_within_tol
            self.last_res_values = res_values
            self.last_residue_values = err_values

            err_values[~self.mask_output] = 0

            if np.all(targets_within_tol | (~self.mask_output)):
                if self.zero_if_met:
                    err_values *= 0
                self.last_point_within_tol = True
                self.found_point_within_tol = True
                if self.verbose:
                    _print("Found point within tolerance!")
            else:
                self.last_point_within_tol = False

            # handle optimize logarithm
            err_values = err_values.copy()
            for ii, tt in enumerate(self.targets):
                if self.mask_output[ii] and tt.optimize_log:
                    assert (
                        res_values[ii] > 0
                    ), "Cannot use optimize_log with negative values"
                    assert tt.value > 0, "Cannot use optimize_log with negative targets"
                    if hasattr(tt, "transform"):
                        assert (
                            tt.transform(res_values[ii]) == res_values[ii]
                        ), "Cannot use optimize_log with transformed values"
                    vvv = np.log10(res_values[ii])
                    vvv_tar = np.log10(tt.value)
                    err_values[ii] = vvv - vvv_tar

            for ii, tt in enumerate(self.targets):
                if tt.weight is not None:
                    err_values[ii] *= tt.weight

        if return_scalar is None:
            return_scalar = self.return_scalar

        if return_scalar:
            out = np.sum(err_values * err_values)
        else:
            out = np.array(err_values)

        if self.show_call_counter:
            _print(
                f"Matching: model call n. {self.call_counter} "
                + (f"penalty = {out:.4g}" if return_scalar else "")
                + "              ",
                end="\r",
                flush=True,
            )
        self.call_counter += 1

        return out

    def get_jacobian(self, x, f0=None):
        x = np.array(x).copy()
        steps = self._knobs_to_x(self.steps_for_jacobian)
        assert len(x) == len(steps)
        if f0 is None:
            f0 = self(x)
        if np.isscalar(f0):
            jac = np.zeros((1, len(x)))
        else:
            jac = np.zeros((len(f0), len(x)))
        mask_input = self.mask_input
        for ii in range(len(x)):
            if not mask_input[ii]:
                continue
            x[ii] += steps[ii]
            jac[:, ii] = (self(x, check_limits=False) - f0) / steps[ii]
            x[ii] -= steps[ii]

        self._last_jac = jac
        return jac

    def _clip_to_max_steps(self, x_step):
        max_steps = np.array([vv.max_step for vv in self.vary])
        out = x_step.copy()
        for ii in range(len(x_step)):
            if max_steps[ii] is None:
                continue
            if np.abs(x_step[ii]) > max_steps[ii]:
                out *= max_steps[ii] / np.abs(out[ii])
        return out

    @property
    def mask_input(self):
        mask = []
        for vv in self.vary:
            if hasattr(vv, "active"):
                mask.append(vv.active)
            else:
                mask.append(True)
        return np.array(mask)

    @property
    def mask_output(self):
        mask = []
        for tt in self.targets:
            if hasattr(tt, "active"):
                mask.append(tt.active)
            else:
                mask.append(True)
        return np.array(mask)


class MeritFuctionView:

    def __init__(self, merit_function, check_limits=True, return_scalar=None):

        self.merit_function = merit_function
        self.check_limits = check_limits
        self.return_scalar = return_scalar

    def __call__(self, x):
        return self.merit_function(
            x, check_limits=self.check_limits, return_scalar=self.return_scalar
        )

    def get_x_limits(self):
        return self.merit_function._get_x_limits()

    def get_x(self):
        return self.merit_function._get_x()

    def set_x(self, x):
        self.merit_function._set_x(x)


class Optimize:

    def __init__(
        self,
        vary,
        targets,
        restore_if_fail=True,
        solver=None,
        verbose=False,
        assert_within_tol=True,
        n_steps_max=20,
        solver_options={},
        show_call_counter=True,
        check_limits=True,
        **kwargs,
    ):
        """
        Numerical optimizer for matching.

        Parameters
        ----------
        vary : list of Vary
            List of knobs to vary.
        targets : list of Target
            List of targets to match.
        restore_if_fail : bool, optional
            If True, restore the initial knob values if the optimization fails.
            Defaults to True.
        solver : str, optional
            Solver to use. Can be 'fsolve', 'bfgs', or 'jacobian'. Defaults to
            'jacobian'.
        verbose : bool, optional
            If True, print information during the optimization. Defaults to False.
        assert_within_tol : bool, optional
            If True, raise an error if the optimization fails. Defaults to True.
        n_steps_max : int, optional
            Maximum number of steps to take. Defaults to 20.
        solver_options : dict, optional
            Options to pass to the solver. Defaults to {}.

        """

        if isinstance(vary, (str, Vary)):
            vary = [vary]

        input_vary = vary
        vary = []
        for ii, rr in enumerate(input_vary):
            if isinstance(rr, Vary):
                vary.append(rr)
            elif isinstance(rr, str):
                vary.append(Vary(rr))
            elif isinstance(rr, (list, tuple)):
                raise ValueError("Not supported")
            elif isinstance(rr, VaryList):
                vary += rr.vary_objects
            else:
                raise ValueError(f"Invalid vary setting {rr}")

        input_targets = targets
        targets = []
        for ii, tt in enumerate(input_targets):
            if isinstance(tt, Target):
                targets.append(tt)
            elif isinstance(tt, (list, tuple)):
                targets.append(Target(*tt))
            elif isinstance(tt, TargetList):
                targets += tt.targets
            else:
                raise ValueError(f"Invalid target element {tt}")

        actions = []
        for tt in targets:
            if tt.weight is None:
                tt.weight = 1.0
            if tt.weight <= 0:
                raise ValueError("`weight` must be positive.")

            if tt.action not in actions:
                actions.append(tt.action)

        for aa in actions:
            aa.prepare()

        data0 = {}
        for aa in actions:
            data0[aa] = aa.run()
            if isinstance(data0[aa], str):
                assert (
                    data0[aa] != "failed"
                ), f"Action {aa} failed to compute initial data."

        for tt in targets:
            if tt.value == "preserve":
                tt.value = tt.eval(data0)

        if solver is None:
            solver = "jacobian"

        if verbose:
            _print(f"Using solver {solver}")

        steps = []
        for vv in vary:
            if vv.step is None:
                steps.append(STEP_DEFAULT)
            else:
                steps.append(vv.step)

        assert solver in ["fsolve", "bfgs", "jacobian"], f"Invalid solver {solver}."

        return_scalar = {"fsolve": False, "bfgs": True, "jacobian": False}[solver]

        _err = MeritFunctionForMatch(
            vary=vary,
            targets=targets,
            actions=actions,
            return_scalar=return_scalar,
            call_counter=0,
            verbose=verbose,
            tw_kwargs=kwargs,
            steps_for_jacobian=steps,
            check_limits=check_limits,
            show_call_counter=show_call_counter,
        )

        if solver == "jacobian":
            self.solver = JacobianSolver(func=_err, verbose=verbose, **solver_options)
        else:
            raise NotImplementedError(f"Solver {solver} not implemented.")

        self.assert_within_tol = assert_within_tol
        self.restore_if_fail = restore_if_fail
        self.n_steps_max = n_steps_max
        self._log = dict(
            penalty=[],
            hit_limits=[],
            alpha=[],
            tol_met=[],
            knobs=[],
            targets=[],
            vary_active=[],
            target_active=[],
            tag=[],
        )

        if not self.check_limits:
            self.add_point_to_log()
            self._clip_to_limits()

        self.add_point_to_log()

    def step(
        self,
        n_steps=1,
        enable_target=None,
        enable_vary=None,
        enable_vary_name=None,
        disable_target=None,
        disable_vary=None,
        disable_vary_name=None,
    ):
        """
        Perform one or more optimization steps.

        Parameters
        ----------
        n_steps : int, optional
            Number of steps to perform. Defaults to 1.
        enable_target: list of int or strings, optional
            For the performed steps, enable target with corresponding id or tag
        enable_vary: list of int or strings, optional
            For the performed steps, enable variables with corresponding id or tag
        enable_vary_name: list of str, optional
            For the performed steps, enable variables with corresponding name
        disable_target: list of int or strings, optional
            For the performed steps, disable target with corresponding id or tag
        disable_vary: list of int or strings, optional
            For the performed steps, disable variables with corresponding id or tag
        disable_vary_name: list of str, optional
            For the performed steps, disable variables with corresponding name
        """
        if not self.check_limits:
            self._clip_to_limits()

        if enable_target is not None:
            self.enable(target=enable_target)

        if enable_vary is not None:
            self.enable(vary=enable_vary)

        if disable_target is not None:
            self.disable(targets=disable_target)

        if disable_vary is not None:
            self.disable(vary=disable_vary)

        if disable_vary_name is not None:
            self.disable(vary_name=disable_vary_name)

        if enable_vary_name is not None:
            self.enable(vary_name=enable_vary_name)

        for i_step in range(n_steps):
            knobs_before = self._extract_knob_values()

            x = self._err._knobs_to_x(knobs_before)
            mskinp = self._err.mask_input
            if self.solver.x is None or not np.allclose(
                x[mskinp], self.solver.x[mskinp], rtol=0, atol=1e-12
            ):
                self.solver.x = x  # this resets solver.mask_from_limits

            # self.solver.x = self._err._knobs_to_x(self._extract_knob_values())
            self.solver.step()
            self._log["penalty"].append(self.solver.penalty_after_last_step)

            self.set_knobs_from_x(self.solver.x)

            knobs_after = self._extract_knob_values()
            self._log["knobs"].append(knobs_after)
            self._log["targets"].append(self._err.last_res_values)
            self._log["hit_limits"].append(
                _bool_array_to_string(~self.solver.mask_from_limits)
            )
            self._log["vary_active"].append(_bool_array_to_string(self._err.mask_input))
            self._log["target_active"].append(
                _bool_array_to_string(self._err.mask_output)
            )
            self._log["tol_met"].append(
                _bool_array_to_string(self._err.last_targets_within_tol)
            )
            self._log["alpha"].append(self.solver.alpha_last_step)
            self._log["tag"].append("")

            if self._err.last_point_within_tol:
                break

        if enable_target is not None:
            self.disable(target=enable_target)

        if enable_vary is not None:
            self.disable(vary=enable_vary)

        if disable_target is not None:
            self.enable(targets=disable_target)

        if disable_vary is not None:
            self.enable(vary=disable_vary)

        if disable_vary_name is not None:
            self.enable(vary_name=disable_vary_name)

        if enable_vary_name is not None:
            self.disable(vary_name=enable_vary_name)

        return self

    def solve(self):
        """
        Perform the optimization, i.e. performs the required number of steps (up
        to `n_steps_max`) to find a point within tolerance.
        If `assert_within_tol` is True, raises an error if no point within
        tolerance is found. If `restore_if_fail` is True, restores the initial
        knob values if no point within tolerance is found.
        """

        try:
            self.solver.x = self._err._knobs_to_x(self._extract_knob_values())
            for ii in range(self.n_steps_max):
                self.step()
                if self.solver.stopped is not None:
                    break

            if not self._err.last_point_within_tol:
                _print("\n")
                _print("Could not find point within tolerance.")

            if self.assert_within_tol and not self._err.last_point_within_tol:
                raise RuntimeError("Could not find point within tolerance.")

            self.set_knobs_from_x(self.solver.x)

        except Exception as err:
            if self.restore_if_fail:
                self.reload(iteration=0)
            _print("\n")
            raise err
        if self._err.show_call_counter:
            _print("\n")

    def vary_status(self, ret=False, max_col_width=40, iter_ref=0):
        """
        Display the status of the knobs.

        Parameters
        ----------
        ret : bool, optional
            If True, return the status as a Table. Defaults to False.
        max_col_width : int, optional
            Maximum column width. Defaults to 40.
        iter_ref : int, optional
            Iteration to use as reference. Defaults to 0.
        """

        vvv = self._vary_table()
        vvv["name"] = np.array([vv.name for vv in self.vary])
        vvv["current_val"] = np.array(self._err._extract_knob_values())
        vvv["lower_limit"] = np.array(
            [(vv.limits[0] if vv.limits is not None else None) for vv in self.vary]
        )
        vvv["upper_limit"] = np.array(
            [(vv.limits[1] if vv.limits is not None else None) for vv in self.vary]
        )
        vvv[f"val_at_iter_{iter_ref}"] = self.log().vary[iter_ref, :]
        vvv["step"] = np.array([vv.step for vv in self.vary])
        vvv["weight"] = np.array([vv.weight for vv in self.vary])

        # check if variable is in limits
        in_lim = []
        for vv, cv, lo, hi in zip(
            self.vary, vvv["current_val"], vvv["lower_limit"], vvv["upper_limit"]
        ):
            good = "OK"
            if lo is not None and cv < lo:
                good = "LOW"
            if hi is not None and cv > hi:
                good = "HIGH"
            in_lim.append(good)
        vvv["met"] = np.array(in_lim)

        vvv._col_names = [
            "id",
            "state",
            "tag",
            "met",
            "name",
            "lower_limit",
            "current_val",
            "upper_limit",
            f"val_at_iter_{iter_ref}",
            "step",
            "weight",
        ]


        if ret:
            return vvv
        else:
            print("Vary status:                 ")
            vvv.show(max_col_width=max_col_width, maxwidth=1000)


    def target_status(self, ret=False, max_col_width=40):
        """
        Display the status of the targets.

        Parameters
        ----------
        ret : bool, optional
            If True, return the status as a Table. Defaults to False.
        max_col_width : int, optional
            Maximum column width. Defaults to 40.
        """

        ttt = self._targets_table()
        self._err(None, check_limits=False)
        ttt["tol_met"] = self._err.last_targets_within_tol
        ttt["residue"] = self._err.last_residue_values
        ttt["current_val"] = np.array(self._err.last_res_values)

        ttt["target_val"] = np.array([tt.value for tt in self.targets])
        ttt._col_names = [
            "id",
            "state",
            "tag",
            "tol_met",
            "residue",
            "current_val",
            "target_val",
            "description",
        ]


        if ret:
            return ttt
        else:
            print("Target status:               ")
            ttt.show(max_col_width=max_col_width, maxwidth=1000)


    def get_knob_values(self, iteration=None):
        """
        Get the knob values at a given iteration.

        Parameters
        ----------
        iteration : int, optional
            Iteration to use. Defaults to None, i.e. the last iteration.

        Returns
        -------
        dict
            Dictionary of knob values.
        """

        if iteration is None:
            iteration = len(self._log["penalty"]) - 1
        out = dict()
        for ii, vv in enumerate(self.vary):
            out[vv.name] = self._log["knobs"][iteration][ii]

        return out

    def show(self, vary=True, targets=True, maxwidth=1000, max_col_width=80):
        """
        Display the knobs and targets used in the optimization.

        Parameters
        ----------
        vary : bool, optional
            If True, display the knobs. Defaults to True.
        targets : bool, optional
            If True, display the targets. Defaults to True.
        maxwidth : int, optional
            Maximum width of the table. Defaults to 1000.
        max_col_width : int, optional
            Maximum column width. Defaults to 80.
        """

        if vary:
            print("Vary:")
            self._vary_table().show(maxwidth=maxwidth, max_col_width=max_col_width)
        if targets:
            print("Targets:")
            self._targets_table().show(maxwidth=maxwidth, max_col_width=max_col_width)

    def log(self):
        """
        Return the optimization log as a Table.

        Returns
        -------
        Table
            Optimization log.
        """

        out_dct = dict()
        out_dct["penalty"] = np.array(self._log["penalty"])
        out_dct["alpha"] = np.array(self._log["alpha"])
        out_dct["tag"] = np.array(self._log["tag"])
        out_dct["tol_met"] = np.array(self._log["tol_met"])
        out_dct["target_active"] = np.array(self._log["target_active"])
        out_dct["hit_limits"] = np.array(self._log["hit_limits"])
        out_dct["vary_active"] = np.array(self._log["vary_active"])
        out_dct["iteration"] = np.arange(len(out_dct["penalty"]))

        knob_array = np.array(self._log["knobs"])
        for ii, vv in enumerate(self.vary):
            out_dct[f"vary_{ii}"] = knob_array[:, ii]

        target_array = np.array(self._log["targets"])
        for ii, tt in enumerate(self.targets):
            out_dct[f"target_{ii}"] = target_array[:, ii]

        out_dct["vary"] = knob_array
        out_dct["targets"] = target_array

        out = Table(out_dct, index="iteration")
        return out

    def reload(self, iteration=None, tag=None):
        """
        Reload the knob values from a given iteration in the optimization log.

        Parameters
        ----------
        iteration : int
            Iteration to use.
        """
        assert iteration is not None or tag is not None
        if tag is not None:
            assert iteration is None
            if tag not in self._log["tag"]:
                raise ValueError(f"Tag `{tag}` not found.")
            iteration = np.where(np.array(self._log["tag"]) == tag)[0][-1]

        assert iteration < len(self._log["penalty"])
        knob_values = self._log["knobs"][iteration]
        mask_input = _bool_array_from_string(self._log["vary_active"][iteration])
        for vv, rr, aa in zip(self.vary, knob_values, mask_input):
            vv.container[vv.name] = rr
            vv.active = aa
        mask_output = _bool_array_from_string(self._log["target_active"][iteration])
        for tt, aa in zip(self.targets, mask_output):
            tt.active = aa
        self.add_point_to_log()

    def clear_log(self):
        """
        Clear the optimization log.
        """

        for kk in self._log:
            self._log[kk].clear()
        self.add_point_to_log()

    def add_point_to_log(self, tag=""):
        """
        Add the current point to the optimization log.

        Parameters
        ----------
        tag : str, optional
            Tag to add to the point. Defaults to ''.
        """

        knobs = self._extract_knob_values()
        self._log["knobs"].append(knobs)
        x = self._err._knobs_to_x(knobs)
        _, penalty = self.solver.eval(x)
        self._log["targets"].append(self._err.last_res_values)
        self._log["penalty"].append(penalty)
        self._log["tol_met"].append(
            _bool_array_to_string(self._err.last_targets_within_tol)
        )
        self._log["hit_limits"].append("".join(["n"] * len(knobs)))
        self._log["vary_active"].append(_bool_array_to_string(self._err.mask_input))
        self._log["target_active"].append(_bool_array_to_string(self._err.mask_output))
        self._log["alpha"].append(-1)
        self._log["tag"].append(tag)

    def tag(self, tag=""):
        """
        Tag the current point in the optimization log.

        Parameters
        ----------
        tag : str, optional
            Tag to add to the point. Defaults to ''.

        """
        self.add_point_to_log(tag=tag)

    def enable(self, target=None, vary=None, vary_name=None):
        """
        Enable a list of variables and targets.

        Parameters
        ----------
        target: str, int, list of int or string, True, False.
            If target is True, enable all targets, if False, disable all targets,
            else enable the targets with corresponding id if target is int or tag if target is str.
            String are matched as regular expression.
        vary: list of int or string
            If True, enable all variables. If False, disable all variables.
            Else enable the variables with corresponding id or tag or all if True.
            String are matched as regular expression.
        vary_name: list of str
            Enable the variables with corresponding name.
            String are matched as regular expression.
        """
        _set_state(self.targets, True, target)
        _set_state(self.vary, True, vary, attr="tag")
        _set_state(self.vary, True, vary_name, attr="name")
        return self

    def disable(self, target=None, vary=None, vary_name=None):
        """
        Disable a list of variables and targets.

        Parameters
        ----------
        target: list of int or string
            If True, disable all targets. If False, enable all targets.
            Else enable the targets with corresponding id or tag or all if True.
            String are matched as regular expression.
        vary: list of int or string
            If True, disable all variables. If False, enable all variables.
            Else enable the variables with corresponding id or tag or all if True.
            String are matched as regular expression.
        vary_name: list of str
            Disable the variables with corresponding name.
            String are matched as regular expression.
        """
        _set_state(self.targets, False, target)
        _set_state(self.vary, False, vary, attr="tag")
        _set_state(self.vary, False, vary_name, attr="name")
        return self

    def get_merit_function(self, check_limits=True, return_scalar=None):
        """
        Get the merit function that can be used with a different optimizer.

        Parameters
        ----------
        check_limits : bool, optional
            If True, enforce that the knob values are within the limits.
            An error is raised if a knob value is outside the limits.
            Defaults to True.
        return_scalar : bool, optional
            If True, return a scalar value. If False, return an array.
            If None, use the default value for the solver. Defaults to None.
        """

        return self._err.get_merit_function(
            check_limits=check_limits, return_scalar=return_scalar
        )

    def _clip_to_limits(self):
        for vv in self.vary:
            if vv.active:
                cv = vv.get_value()
                if vv.limits is None:
                    continue
                if vv.limits[0] is not None:
                    if cv < vv.limits[0]:
                        vv.container[vv.name] = vv.limits[0]
                if vv.limits[1] is not None:
                    if cv > vv.limits[1]:
                        vv.container[vv.name] = vv.limits[1]

    #### DEPRECATED METHODS ####

    def enable_vary(self, id=None, tag=None):
        """
        Enable one or more variables.

        Deprecated. Please use enable(vary=id_or_tags).

        Parameters
        ----------
        id : int or list of int, optional
            Index of the variable to disable. Defaults to None.
        tag : str or list of str, optional
            Tag of the variable to disable.
        """
        #log.warning(
        #    "WARNING: `enable_vary` will be deprecated."
        #    " Please use `enable(vary=id_or_tags)`."
        #)
        self.enable(vary=_add_id_tag(id, tag))

    def disable_vary(self, id=None, tag=None):
        """
        Disable one or more variables.

        Deprecated. Please use `disable(vary=id_or_tags)`.

        Parameters
        ----------
        id : int or list of int, optional
            Index of the variable to disable. Defaults to None.
        tag : str or list of str, optional
            Tag of the variable to disable.
            Str is interpreted as regular expression. Defaults to None.
        """
        #log.warning(
        #    "WARNING: `disable_vary` will be deprecated."
        #    " Please use `disable(vary=id_or_tags)`."
        #)
        self.disable(vary=_add_id_tag(id, tag))

    def enable_targets(self, *id_or_tag, id=None, tag=None):
        """
        Enable one or more targets.

        Deprecated. Please use `enable(target=id_or_tags)` instead.

        Parameters
        ----------
        id_or_tag : int or string
            Disable the targets with corresponding id or tag.
        Depraecated:
        id : int or list of int, optional
            Index of the targets to disable. Defaults to None.
        tag : str or list of str, optional
            Tag of the targets to disable.
            Str is interpreted as regular expression. Defaults to None.
        """
        #log.warning(
        #    "WARNING: `enable_targets` will be deprecated."
        #    " Please use `enable(target=id_or_tags)`."
        #)

        self.enable(target=_add_id_tag(id, tag))

    def disable_targets(self, id=None, tag=None):
        """
        Disable one or more targets.

        Deprecated. Use `disable(target=id_or_tags)` instead.

        Parameters
        ----------
        id : int or list of int, optional
            Index of the targets to disable. Defaults to None.
        tag : str or list of str, optional
            Tag of the targets to disable.
            Str is interpreted as regular expression. Defaults to None.
        """
        #log.warning(
        #    "WARNING: `disable_targets` will be deprecated."
        #    " Please use `disable(target=ids_or_tags)`."
        #)
        self.disable(target=_add_id_tag(id, tag))

    def disable_all_targets(self):
        """
        Disable all targets.

        Deprecated. Please use `disable(target=True)`.
        """

        #log.warning(
        #    "WARNING: `disable_all_targets` will be deprecated."
        #    " Please use `disable(target=True)."
        #)

        for tt in self.targets:
            tt.active = False

    def enable_all_targets(self):
        """
        Enable all targets.

        Deprecated. Please use `enable(target=True)`.
        """

        #log.warning(
        #    "WARNING: `enable_all_targets` will be deprecated."
        #    " Please use `enable(target=True)."
        #)

        for tt in self.targets:
            tt.active = True
        return self

    def disable_all_vary(self):
        """
        Disable all knobs.

        Deprecated. Please use `disable(vary=True)`.
        """

        #log.warning(
        #    "WARNING: `disable_all_vary` will be deprecated."
        #    " Please use `disable(vary=True)."
        #)

        for vv in self.vary:
            vv.active = False
        return self

    def enable_all_vary(self):
        """
        Enable all knobs.

        Deprecated. Please use `enable(vary=True)`.
        """

        #log.warning(
        #    "WARNING: `enable_all_vary` will be deprecated."
        #    " Please use `enable(vary=True)."
        #)

        for vv in self.vary:
            vv.active = True
        return self

    #### END DEPRECATED METHODS ####

    @property
    def check_limits(self):
        return self._err.check_limits

    @check_limits.setter
    def check_limits(self, value):
        self._err.check_limits = value

    @property
    def _err(self):
        return self.solver.func

    @property
    def actions(self):
        return self._err.actions

    @property
    def verbose(self):
        return self.solver.verbose

    @verbose.setter
    def verbose(self, value):
        self.solver.verbose = value

    @property
    def vary(self):
        return self._err.vary

    @property
    def targets(self):
        return self._err.targets

    def _is_within_tol(self):
        return self._err()

    def set_knobs_from_x(self, x):
        for vv, rr in zip(self.vary, self._err._x_to_knobs(x)):
            if vv.active:
                vv.container[vv.name] = rr

    def _vary_table(self):
        return _make_table(self.vary)

    def _targets_table(self):
        return _make_table(self.targets)

    def _extract_knob_values(self):
        return self._err._extract_knob_values()


def _bool_array_to_string(arr, dct={True: "y", False: "n"}):
    return "".join([dct[aa] for aa in arr])


def _bool_array_from_string(strng, dct={"y": True, "n": False}):
    for ss in strng:
        assert ss in dct, f"Invalid character {ss}"
    return np.array([dct[ss] for ss in strng])


def _make_table(vary):
    id = []
    tag = []
    state = []
    description = []
    for ii, vv in enumerate(vary):
        id.append(ii)
        tag.append(vv.tag)
        state.append("ON" if vv.active else "OFF")
        vv_repr = vv.__repr__()
        vv_repr = vv_repr.replace("Vary(", "")
        vv_repr = vv_repr.replace("Vary", "")
        vv_repr = vv_repr.replace("TargetPhaseAdv(", "")
        vv_repr = vv_repr.replace("Target(", "")
        vv_repr = vv_repr.replace("Target", "")
        if vv_repr[-1] == ")":
            vv_repr = vv_repr[:-1]
        description.append(vv_repr)
    id = np.array(id)
    tag = np.array(tag)
    state = np.array(state)
    description = np.array(description)
    return Table(dict(id=id, tag=tag, state=state, description=description), index="id")


def _set_state(lst, state, entries, attr="tag"):
    if entries is None:
        return
    elif entries is True:
        for vv in lst:
            vv.active = state
        return
    elif entries is False:
        for vv in lst:
            vv.active = not state
        return
    if isinstance(entries, int) or isinstance(entries, str):
        entries = [entries]
    for entry in entries:
            if isinstance(entry, int):
                lst[entry].active = state
            elif isinstance(entry, str):
                for vv in lst:
                    if re.fullmatch(entry, getattr(vv, attr)):
                        vv.active = state


def _add_id_tag(id, tag):
    id_or_tag = ()
    if isinstance(id, int):
        id_or_tag += (id,)
    elif isinstance(id, (list, tuple, np.ndarray)):
        id_or_tag += tuple(id)
    if isinstance(tag, str):
        id_or_tag += (tag,)
    elif isinstance(tag, (list, tuple, np.ndarray)):
        id_or_tag += tuple(tag)
    return id_or_tag



import builtins
import math
import operator

BINARY_OPS = {
    operator.add: "+",
    operator.sub: "-",
    operator.mul: "*",
    operator.matmul: "@",
    operator.truediv: "/",
    operator.floordiv: "//",
    operator.mod: "%",
    operator.pow: "**",
    operator.and_: "&",
    operator.or_: "|",
    operator.xor: "^",
    operator.lt: "<",
    operator.le: "<=",
    operator.ne: "!=",
    operator.ge: ">=",
    operator.gt: ">",
    operator.rshift: ">>",
    operator.lshift: "<<",
}

UNARY_OPS = {
    operator.neg: "-",
    operator.pos: "+",
    operator.invert: "~",
}

GUARD_ZERO_DIVISION = ['__truediv__', '__floordiv__', '__mod__']


def generate_binary_refs():
    for op, op_str in BINARY_OPS.items():
        op_name = op.__name__
        if op_name == 'and_':
            op_name = 'BitwiseAnd'
        elif op_name == 'or_':
            op_name = 'BitwiseOr'
        else:
            op_name = op_name.title()
        print(f"""\
@cython.cclass
class {op_name}Expr(BinOpExpr):
    _op_str = '{op_str}'

    def _get_value(self):
        lhs = ARef._mk_value(self._lhs)
        rhs = ARef._mk_value(self._rhs)
""")
        if op_name not in GUARD_ZERO_DIVISION:
            print(f"""\
        return lhs {op_str} rhs
""")
        else:
            print(f"""\
        try:
            return lhs {op_str} rhs
        except ZeroDivisionError:
            return float('nan')
""")


def generate_unary_refs():
    for op, op_str in UNARY_OPS.items():
        print(f"""\
@cython.cclass
class {op.__name__.title()}Expr(UnaryOpExpr):
    _op_str = '{op_str}'
    
    def _get_value(self):
        arg = ARef._mk_value(self._arg)
        return {op_str}arg
        
""")


generate_binary_refs()
generate_unary_refs()



import numpy as np
import pytest

from xdeps import Table

data = {
    "name": np.array(["ip1", "ip2", "ip2", "ip3", "tab$end"]),
    "s": np.array([1.0, 2.0, 2.1, 3.0, 4.0]),
    "betx": np.array([4.0, 5.0, 5.1, 6.0, 7.0]),
    "bety": np.array([2.0, 3.0, 3.1, 4.0, 9.0]),
}


## Table tests

t = Table(data)

def test_table_initialization():
    # Valid initialization
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)
    assert len(table) == 3
    assert table._col_names == ["name", "value"]

    # Invalid initialization (non-numpy array)
    data_invalid = {"name": ["a", "b", "c"], "value": [1, 2, 3]}
    try:
        Table(data_invalid)
    except ValueError as e:
        assert str(e) == "Column `name` is not a numpy array"

    # Invalid initialization (different column lengths)
    data_invalid_length = {"name": np.array(["a", "b"]), "value": np.array([1, 2, 3])}
    try:
        Table(data_invalid_length)
    except ValueError as e:
        assert str(e) == "Columns have different lengths"


def test_table_split_names():
    assert t._split_name_count_offset("example") == ("example", None, 0)
    assert t._split_name_count_offset("example::5") == ("example", 5, 0)
    assert t._split_name_count_offset("example<<3") == ("example", None, -3)
    assert t._split_name_count_offset("example::5<<3") == ("example", 5, -3)
    assert t._split_name_count_offset("example::5<<-3") == ("example", 5, 3)
    assert t._split_name_count_offset("example<<-3") == ("example", None, 3)
    assert t._split_name_count_offset("example::10<<0") == ("example", 10, 0)
    assert t._split_name_count_offset("example::0<<-1") == ("example", 0, 1)
    assert t._split_name_count_offset("example<<0") == ("example", None, 0)
    assert t._split_name_count_offset("example::0<<0") == ("example", 0, 0)
    assert t._split_name_count_offset("example>>3") == ("example", None, 3)
    assert t._split_name_count_offset("example::5>>3") == ("example", 5, 3)
    assert t._split_name_count_offset("example::5>>-3") == ("example", 5, -3)
    assert t._split_name_count_offset("example>>-3") == ("example", None, -3)
    assert t._split_name_count_offset("example::10>>0") == ("example", 10, 0)
    assert t._split_name_count_offset("example::0>>-1") == ("example", 0, -1)
    assert t._split_name_count_offset("example>>0") == ("example", None, 0)
    assert t._split_name_count_offset("example::0>>0") == ("example", 0, 0)


def test_table_len():
    assert len(t.betx) == len(data["betx"])


def test_table_getitem_col():
    assert id(t["betx"]) == id(data["betx"])
    assert t["betx+sqrt(bety)"][1] == (t.betx + np.sqrt(t.bety))[1]
    try:
        t["bbb"]
    except NameError as e:
        assert str(e) == "name 'bbb' is not defined"


def test_table_getitem_col_row():
    assert t["betx", 0] == data["betx"][0]
    assert t["betx", "ip1"] == data["betx"][0]
    assert t["betx", "ip2"] == data["betx"][1]
    assert t["betx", ("ip2", 0)] == data["betx"][1]
    assert t["betx", ("ip2", 1)] == data["betx"][2]
    assert t["betx", ("ip2",-1)] == data["betx"][2]
    assert t["betx", "ip2::0"] == data["betx"][1]
    assert t["betx", "ip2::1"] == data["betx"][2]
    assert t["betx", "ip2::-1"] == data["betx"][2]
    assert t["betx", ("ip2", 1)] == data["betx"][2]
    assert t["betx", "ip2<<1"] == data["betx"][0]
    assert t["betx", ("ip2", 0, -1)] == data["betx"][0]
    assert t["betx", "ip2::1>>1"] == data["betx"][3]
    assert t["betx", ("ip2", 1, 1)] == data["betx"][3]
    assert t["betx", "ip2::-1"] == data["betx"][2]
    assert t["betx", ("ip2", -1, 1)] == data["betx"][3]
    assert np.all(t["betx", 0:2] == data["betx"][0:2])
    assert np.all(t["betx", "ip1":"ip3"] == data["betx"][0:4])
    assert np.all(t["betx", "ip1":"ip3":"name"] == data["betx"][0:4])
    assert np.all(t["betx", "ip1":"ip2::1"] == data["betx"][0:3])
    assert np.all(t["betx", ["ip3", "ip2::1"]] == data["betx"][[3, 2]])
    assert np.all(t["betx", [3, 2]] == data["betx"][[3, 2]])
    assert np.all(
        t["betx", [True, False, True, False, True]] == data["betx"][[0, 2, 4]]
    )
    with pytest.raises(KeyError) as e:
        t["betx", "notthere"]
        assert str(e) == "Cannot find 'notthere' in column 'name'"


def test_table_numpy_string():
    tab = Table(dict(name=np.array(["a", "b$b"]), val=np.array([1, 2])))
    assert tab["val", tab.name[1]] == 2


def test_table_column_access():
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)

    # Direct column access
    assert np.array_equal(table["name"], data["name"])
    assert np.array_equal(table["value"], data["value"])

    # Column expression
    assert np.array_equal(table["value * 2"], data["value"] * 2)


def test_table_sort_columns():
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)

    # Sort columns
    assert table._col_names == ["name", "value"]
    table._col_names = ["value", "name"]
    assert table._col_names == ["value", "name"]


def test_table_column_assignment():
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    data_orig = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)

    # Direct column assignment
    table["value"] = table["value"] * 2
    assert np.array_equal(table["value"], data["value"])
    assert np.array_equal(table["value"], data_orig["value"] * 2)

    # Direct column assignment
    table["value"] = 1
    assert np.array_equal(table["value"], np.array([1, 1, 1]))
    assert np.array_equal(data["value"], np.array([1, 1, 1]))


def test_table_del_column():
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)

    # Delete column
    del table["value"]
    assert "value" not in table._col_names


def test_table_from_methods():
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)

    # Show method
    table.show()

    # From pandas
    import pandas as pd

    df = pd.DataFrame(data)
    table_from_pandas = Table.from_pandas(df)
    assert len(table_from_pandas) == 3

    # From CSV
    df.to_csv("test.csv", index=False)
    table_from_csv = Table.from_csv("test.csv")
    assert len(table_from_csv) == 3

    # From rows
    rows = [
        {"name": "a", "value": 1},
        {"name": "b", "value": 2},
        {"name": "c", "value": 3},
    ]
    table_from_rows = Table.from_rows(rows)
    assert len(table_from_rows) == 3

    # To pandas
    df_from_table = table._df
    assert df_from_table.equals(df)


def test_table_concatenate():
    data1 = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    data2 = {"name": np.array(["d", "e", "f"]), "value": np.array([4, 5, 6])}
    table1 = Table(data1)
    table2 = Table(data2)

    # Concatenate
    concatenated_table = table1 + table2
    assert len(concatenated_table) == 6
    assert np.array_equal(concatenated_table["value"], np.array([1, 2, 3, 4, 5, 6]))

def test_table_transpose():
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)

    transposed_table = table._t
    assert transposed_table._col_names == ["columns", "row0", "row1", "row2"]


def test_edge_cases():
    # Empty table
    data_empty = {"name": np.array([]), "value": np.array([])}
    table_empty = Table(data_empty)
    assert len(table_empty) == 0

    # Non-uniform column lengths
    data_non_uniform = {"name": np.array(["a", "b"]), "value": np.array([1, 2, 3])}
    try:
        Table(data_non_uniform)
    except ValueError as e:
        assert str(e) == "Columns have different lengths"


def test_table_show():
    data = {"name": np.array(["a", "b", "c"]), "value": np.array([1, 2, 3])}
    table = Table(data)
    assert table.show(output=str) == "name value\na        1\nb        2\nc        3"


## Table cols tests

def test_cols():
    assert isinstance(t.cols["betx"], Table)
    assert t.cols["betx", "bety"].betx[0] == t.betx[0]

def test_cols_iter():
    data = {
        "name": np.array(["a", "b", "c"]),
        "c1": np.array([1, 2, 3]),
        "c2": np.array([4, 5, 6]),
        "s1":1,
        "s2":2,
    }
    with pytest.raises(ValueError) as err:
        Table(data)
        assert str(err) == "Column `s1` is not a numpy array"

    table = Table(data, col_names=["name", "c1", "c2"])

    for col in table.cols:
        assert col in table._col_names

    assert list(table.cols) == ["name", "c1", "c2"]

def test_cols_get_index_unique():
    assert len(set(t.cols.get_index_unique())) == len(t)

## Table rows tests

def test_rows_get_index():
    assert t.rows.get_index("ip2::1") == 2
    assert t.rows.get_index("ip2") == 1
    assert t.rows.get_index(("ip2",1)) == 2

def test_rows_is_repeated():
    assert not t.rows.is_repeated("ip1")
    assert t.rows.is_repeated("ip2")
    assert not t.rows.is_repeated("ip3")
    assert not t.rows.is_repeated("tab$end")

def test_rows_selection_indices():
    assert t.rows[1].betx[0] == data["betx"][1]
    assert np.array_equal(t.rows[[2, 1]].betx, data["betx"][[2, 1]])


def test_rows_selection_names():
    assert t.rows["ip2"].betx[0] == data["betx"][1]
    assert t.rows["ip[23]"].betx[0] == data["betx"][1]
    assert t.rows["ip.*::1"].betx[0] == data["betx"][2]
    assert len(t.rows["notthere"]) == 0
    assert t.rows[["ip1", "ip2"]].betx[1] == data["betx"][1]


def test_rows_selection_ranges():
    assert t.rows[1:4:3].betx[0] == data["betx"][1]
    assert t.rows[1.5:2.5:"s"].betx[0] == data["betx"][1]
    assert t.rows["ip1":"ip3"].betx[2] == data["betx"][2]
    assert t.rows["ip2::1<<1":"ip2::1>>1"].betx[0] == data["betx"][1]
    assert t.rows["ip1":"ip3":"name"].betx[0] == data["betx"][0]
    assert t.rows[:].betx[0] == data["betx"][0]


def test_rows_multiple_selection():
    assert t.rows[t.s > 1, 1].betx[0] == data["betx"][t.s > 1][1]

def test_rows_mask():
    assert np.array_equal(t.betx[t.rows.mask[:, t.s > 1]], data["betx"][t.s > 1])
    assert np.array_equal(t.betx[t.rows.mask[[2, 1]]], data["betx"][[1, 2]])



## Table without index

def test_table_no_index():
    data = {
        "name": np.array(["a", "b", "c"]),
        "c1": np.array([1, 2, 3]),
        "c2": np.array([4, 5, 6]),
    }
    table = Table(data, index=None)
    assert "name" in table._col_names
    assert "c1" in table._col_names
    assert "c2" in table._col_names
    assert table._index is None

    assert table["c1", 1] == data["c1"][1]
    assert np.array_equal(table["c1"], data["c1"])
    assert np.array_equal(table["c1", [1, 2]], data["c1"][[1,2]])







# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

# cython: language_level=3

import builtins
import cython
import logging
import math
import operator

logger = logging.getLogger(__name__)

special_methods = {
    '__copy__',
    '__deepcopy__',
    '__dict__',
    '__getstate__',
    '__setstate__',
    '__reduce_cython__',
    '__wrapped__',
    '__array_ufunc__',
    '__array_function__',
    '__array_struct__',
    '__array_interface__',
    '__array_prepare__',
    '__array_wrap__',
    '__array_finalize__',
    '__array__',
    '__array_priority__',
    '__iter__',
    # _ipython_canary_method_should_not_exist_ is used by IPython to detect
    # if an object 'lies' about its attributes due to its __getattr__. We should
    # not try to intercept it.
}

OPERATOR_SYMBOLS = {
    # Binary
    operator.add: "+",
    operator.sub: "-",
    operator.mul: "*",
    operator.matmul: "@",
    operator.truediv: "/",
    operator.floordiv: "//",
    operator.mod: "%",
    operator.pow: "**",
    operator.and_: "&",
    operator.or_: "|",
    operator.xor: "^",
    operator.lt: "<",
    operator.le: "<=",
    operator.eq: "==",
    operator.ne: "!=",
    operator.ge: ">=",
    operator.gt: ">",
    operator.rshift: ">>",
    operator.lshift: "<<",
    # Unary
    operator.neg: "-",
    operator.pos: "+",
    operator.invert: "~",
    # Inplace
    operator.iadd: "+",
    operator.isub: "-",
    operator.imul: "*",
    operator.imatmul: "@",
    operator.itruediv: "/",
    operator.ifloordiv: "//",
    operator.imod: "%",
    operator.ipow: "**",
    operator.ilshift: "<<",
    operator.irshift: ">>",
    operator.iand: "&",
    operator.ior: "|",
    operator.ixor: "^",
}


def is_ref(obj):
    return isinstance(obj, BaseRef)


def _isref(obj):
    # This is never used within this module, let's not prefix it with '_'.
    logger.warning("xdeps.refs._isref is deprecated, use is_ref instead.")
    return is_ref(obj)


def is_cythonized():
    return cython.compiled


@cython.cclass
class BaseRef:
    """
    An abstract base class for all reference/expression objects. Such an object
    represents a value computation, based on dependent values/other refs. E.g.,
    a simple reference to a value contained in a Python object (or a dictionary,
    list field), or a more complex expression, such as a binary operation, whose
    arguments can be refs themselves.

    Notes:
        Cannot override __complex__, because Python requires it to return a
        complex value, but we want it to be a Ref object. The same is true for
        __int__, __float__, __bool__. The previous implementations are therefore
        removed, and no workaround is provided until an explicit need for these
        methods arises.
    """
    _manager = cython.declare(object, visibility='readonly', value=None)
    _hash = cython.declare(int, visibility='readonly')

    def __init__(self, *args, **kwargs):
        # To keep compatibility with pure Python (useful for debugging simpler
        # issues), we simulate Cython __cinit__ behaviour with this __init__:
        if not is_cythonized():
            for base in type(self).__mro__:
                cinit = getattr(base, '__cinit__', None)
                if cinit:
                    cinit(self, *args, **kwargs)

    def __hash__(self):
        return self._hash

    def __reduce__(self):
        raise TypeError("Cannot pickle an abstract class")

    def __eq__(self, other):
        """Check equality of the expressions `self` and `other`.

        Check if `self` and `other` are the same. Crucially, even if self._value
        and other._value are the same, this does not necessarily imply that
        self == other. Python requires that if a == b then hash(a) == hash(b).
        Since in our case two different ref objects may have the same hash (by
        virtue of representing the same expression!), we must not use __eq__ to
        create expressions.

        If a deferred equality expression is desired, use `self._eq`.
        """
        return str(self) == str(other)

    def _eq(self, other):
        """Deferred expression for the equality of values of `self` and `other`."""
        return EqExpr(self, other)

    def _neq(self, other):
        """Deferred expression for the inequality of values of `self` and `other`."""
        return NeExpr(self, other)

    @staticmethod
    def _mk_value(value):
        if isinstance(value, BaseRef):
            return value._get_value()
        else:
            return value

    def _get_value(self):
        raise NotImplementedError

    @property
    def _value(self):
        try:
            return self._get_value()
        except AttributeError:
            # Python's default __getattribute__ implementation falls back to
            # __getattr__ if a property getter raises AttributeError. This will
            # have an effect of silencing the error in our special case of
            # ItemRef and AttrRef, so here we reraise it as a different type.
            raise LookupError(
                f"The reference '{self}' points to a non-existent field."
            )

    def _get_dependencies(self, out=None):
        return out or set()

    # order of precedence
    def __call__(self, *args, **kwargs):
        return CallRef(self, args, kwargs)

    def __getitem__(self, item):
        return ItemRef(self, item, self._manager)

    def __getattr__(self, attr):
        if attr in special_methods:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{attr}'")

        return AttrRef(self, attr, self._manager)

    # numerical unary  operator
    def __neg__(self):
        return NegExpr(self)

    def __pos__(self):
        return PosExpr(self)

    def __invert__(self):
        return InvertExpr(self)

    # numerical binary operators
    def __add__(self, other):
        return AddExpr(self, other)

    def __radd__(self, other):
        return AddExpr(other, self)

    def __sub__(self, other):
        return SubExpr(self, other)

    def __rsub__(self, other):
        return SubExpr(other, self)

    def __mul__(self, other):
        return MulExpr(self, other)

    def __rmul__(self, other):
        return MulExpr(other, self)

    def __matmul__(self, other):
        return MatmulExpr(self, other)

    def __rmatmul__(self, other):
        return MatmulExpr(other, self)

    def __truediv__(self, other):
        return TruedivExpr(self, other)

    def __rtruediv__(self, other):
        return TruedivExpr(other, self)

    def __floordiv__(self, other):
        return FloordivExpr(self, other)

    def __rfloordiv__(self, other):
        return FloordivExpr(other, self)

    def __mod__(self, other):
        return ModExpr(self, other)

    def __rmod__(self, other):
        return ModExpr(other, self)

    def __pow__(self, other):
        return PowExpr(self, other)

    def __rpow__(self, other):
        return PowExpr(other, self)

    def __and__(self, other):
        return BitwiseAndExpr(self, other)

    def __rand__(self, other):
        return BitwiseAndExpr(other, self)

    def __or__(self, other):
        return BitwiseOrExpr(self, other)

    def __ror__(self, other):
        return BitwiseOrExpr(other, self)

    def __xor__(self, other):
        return XorExpr(self, other)

    def __rxor__(self, other):
        return XorExpr(other, self)

    def __lt__(self, other):
        return LtExpr(self, other)

    def __rlt__(self, other):
        return LtExpr(other, self)

    def __le__(self, other):
        return LeExpr(self, other)

    def __rle__(self, other):
        return LeExpr(other)

    def __ge__(self, other):
        return GeExpr(self, other)

    def __rge__(self, other):
        return GeExpr(other, self)

    def __gt__(self, other):
        return GtExpr(self, other)

    def __rgt__(self, other):
        return GtExpr(other, self)

    def __rshift__(self, other):
        return RshiftExpr(self, other)

    def __rrshift__(self, other):
        return RshiftExpr(other, self)

    def __lshift__(self, other):
        return LshiftExpr(self, other)

    def __rlshift__(self, other):
        return LshiftExpr(other, self)

    def __divmod__(self, other):
        return BuiltinRef(self, builtins.divmod, (other,))

    def __round__(self, other=0):
        return BuiltinRef(self, builtins.round, (other,))

    def __trunc__(self):
        return BuiltinRef(self, math.trunc)

    def __floor__(self):
        return BuiltinRef(self, math.floor)

    def __ceil__(self):
        return BuiltinRef(self, math.ceil)

    def __abs__(self):
        return BuiltinRef(self, builtins.abs)


@cython.cclass
class MutableRef(BaseRef):
    """
    An (abstract) class representing a reference to a mutable object.
    """
    _owner = cython.declare(object, visibility='readonly', value=None)
    _key = cython.declare(object, visibility='readonly', value=None)

    def __cinit__(self, _owner, _key, _manager):
        self._owner = _owner
        self._key = _key
        self._manager = _manager
        # _hash will depend on the particularities of the subclass, for now
        # it is None, which does not matter, as this class should never be
        # instantiated.

    def __setitem__(self, key, value):
        ref = ItemRef(self, key, self._manager)
        self._manager.set_value(ref, value)

    def __setattr__(self, attr, value):
        """Set a built-in attribute of the object or create an AttrRef.

        This should only be called during __init__, when unpickling, or to
        create an AttrRef for attrs that are not already defined on the object.
        In other cases the behaviour between python and cython is different,
        due to the way cython handles __setattr__.

        For this method to work correctly, all subclasses of MutableRef must
        be cython cdef classes (decorated with @cython.cclass).
        """
        if attr in dir(self):
            if not cython.compiled:
                object.__setattr__(self, attr, value)
                return
            else:
                # The above way of setting attributes does not work in Cython,
                # as the object does not have a __dict__. We do not really need
                # a setter for those though, as the only time we need to
                # set a "built-in" attribute is during __cinit__ or when
                # unpickling, and both of those cases are handled by Cython
                # without the usual pythonic call to __setattr__.
                raise AttributeError(f"Attribute {attr} is read-only.")

        ref = AttrRef(self, attr, self._manager)
        self._manager.set_value(ref, value)

    def __reduce__(self):
        """Do not store the hash when pickling.

        The hash is only guaranteed to be the same for the 'same' refs within
        the same python instance, therefore serialising hashes makes no sense.
        """
        return type(self), (self._owner, self._key, self._manager)

    def _get_dependencies(self, out=None):
        if out is None:
            out = set()
        if isinstance(self._owner, BaseRef):
            self._owner._get_dependencies(out)
        if isinstance(self._key, BaseRef):
            self._key._get_dependencies(out)
        out.add(self)
        return out

    def _eval(self, expr, gbl=None):
        if gbl is None:
            gbl = {}
        return eval(expr, gbl, self)  # noqa

    def _exec(self, expr, gbl=None):
        if gbl is None:
            gbl = {}
        exec(expr, gbl, self)  # noqa

    @property
    def _tasks(self):
        """All tasks that influence this expression."""
        return self._manager.tartasks[self]

    def _find_dependant_targets(self):
        return self._manager.find_deps([self])

    @property
    def _expr(self):
        if self in self._manager.tasks:
            task = self._manager.tasks[self]
            if hasattr(task, "expr"):
                return task.expr

    def _info(self, limit=10):
        try:
            value = self._get_value()
            print(f"#  {self}._get_value()")
            print(f"   {self} = {value}")
        except AttributeError:
            print(f"#  The field '{self}' does not exist!!!")
        except NotImplementedError:
            print(f"#  '{self}' does not implement _get_value()!!!")
        print()

        if self in self._manager.tasks:
            task = self._manager.tasks[self]
            print(f"#  {self}._expr")
            print(f"   {task}")
            print()
            if hasattr(task, "info"):
                task.info()
        else:
            print(f"#  {self}._expr is None")
            print()

        refs = self._manager.find_deps([self])[1:]
        limit = limit or len(refs)
        if len(refs) == 0:
            print(f"#  {self} does not influence any target")
            print()
        else:
            print(f"#  {self}._find_dependant_targets()")
            for tt in refs[:limit]:
                if tt._expr is not None:
                    print(f"   {tt}")
            if len(refs) > limit:
                print("   ... set _info(limit=None) to get all lines")
            print()

    def _info2(self, limit=10):
        print(f"Info of {self}")
        print()
        try:
            value = self._get_value()
            print(f"Value: {value}")
        except AttributeError:
            print(f"  The field '{self}' does not exist!!!")
        except NotImplementedError:
            print(f"  '{self}' does not implement _get_value()!!!")
        print()

        if self in self._manager.tasks:
            task = self._manager.tasks[self]
            print(f"Is updated by: {task}")
            print("  where:")
            for expr in task.expr._get_dependencies():
                print(f"  {expr} = {expr._get_value()}")
            print()

        refs = self._manager.find_deps([self])[1:]
        limit = limit or len(refs)
        if len(refs) == 0:
            print("Does not influence any target")
            print()
        else:
            print("Influences: ")
            for tt in refs[:limit]:
                if tt._expr is not None:
                    print(f"   {tt}")
            if len(refs) > limit:
                print("  ... set _info(limit=None) to get all lines")
            print()

    def __iadd__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr + other
        else:
            return self._get_value() + other

    def __ifloordiv__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr // other
        else:
            return self._get_value() // other

    def __ilshift__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr << other
        else:
            return self._get_value() << other

    def __imatmul__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr @ other
        else:
            return self._get_value() @ other

    def __imod__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr % other
        else:
            return self._get_value() % other

    def __imul__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr * other
        else:
            return self._get_value() * other

    def __ipow__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr**other
        else:
            return self._get_value() ** other

    def __irshift__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr >> other
        else:
            return self._get_value() >> other

    def __isub__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr - other
        else:
            return self._get_value() - other

    def __itruediv__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr / other
        else:
            return self._get_value() / other

    def __ixor__(self, other):
        newexpr = self._expr
        if newexpr:
            return newexpr ^ other
        else:
            return self._get_value() ^ other


@cython.cclass
class Ref(MutableRef):
    """
    A reference in the top-level container.
    """
    def __cinit__(self, _owner, _key, _manager):
        # Cython automatically calls __cinit__ in the base classes
        self._hash = hash((type(self).__name__, _key))

    def __repr__(self):
        return self._key

    def _get_value(self):
        return BaseRef._mk_value(self._owner)

    def _get_dependencies(self, out=None):
        return out


@cython.cclass
class ObjectAttrRef(Ref):
    """Like `Ref`, but translates attribute access to item access."""

    def __getattr__(self, attr):
        if attr in special_methods:
            raise AttributeError(
                f"'{type(self).__name__}' object has no attribute '{attr}'")

        return ItemRef(self, attr, self._manager)

    def __setattr__(self, attr, value):
        """Set a built-in attribute of the object or create an ItemRef.

        For notes on why the implementation is different in Cython, see
        the __setattr__ method of `Ref`.
        """
        if attr in dir(self):
            if not cython.compiled:
                object.__setattr__(self, attr, value)
                return
            else:
                raise AttributeError(f"Attribute {attr} is read-only.")

        ref = ItemRef(self, attr, self._manager)
        self._manager.set_value(ref, value)


@cython.cclass
class AttrRef(MutableRef):
    def __cinit__(self, _owner, _key, _manager):
        # Cython automatically calls __cinit__ in the base classes
        self._hash = hash((type(self).__name__, _owner, _key))

    def _get_value(self):
        owner = BaseRef._mk_value(self._owner)
        attr = BaseRef._mk_value(self._key)
        return getattr(owner, attr)

    def _set_value(self, value):
        owner = BaseRef._mk_value(self._owner)
        attr = BaseRef._mk_value(self._key)
        setattr(owner, attr, value)

    def __repr__(self):
        assert self._owner is not None
        assert self._key is not None
        return f"{self._owner}.{self._key}"


@cython.cclass
class ItemRef(MutableRef):
    def __cinit__(self, _owner, _key, _manager):
        # Cython automatically calls __cinit__ in the base classes
        self._hash = hash((type(self).__name__, _owner, _key))

    def _get_value(self):
        owner = BaseRef._mk_value(self._owner)
        item = BaseRef._mk_value(self._key)
        return owner[item]

    def _set_value(self, value):
        owner = BaseRef._mk_value(self._owner)
        item = BaseRef._mk_value(self._key)
        owner[item] = value

    def __repr__(self):
        assert self._owner is not None
        return f"{self._owner}[{repr(self._key)}]"


@cython.cclass
class BinOpExpr(BaseRef):
    """
    This _abstract_ class represents a binary operation on two refs/literals.

    When overriden, the subclass must define the following:

    1. A method `_get_value` which returns the value of the expression. Tests
       have determined that this way of implementing operators (by specialising)
       classes instead of using the generic `operator` methods is ~15% faster
       in both evaluating the expressions and building the dependency graph.
    2. A property `_op_str` which returns the string representation of the
       operator for pretty printing.
    """
    _lhs = cython.declare(object, visibility='readonly')
    _rhs = cython.declare(object, visibility='readonly')

    def __cinit__(self, lhs, rhs):
        self._lhs = lhs
        self._rhs = rhs
        self._hash = hash((self.__class__, lhs, rhs))

    def _get_value(self):
        raise NotImplementedError()

    def _get_dependencies(self, out=None):
        if out is None:
            out = set()
        if isinstance(self._lhs, BaseRef):
            self._lhs._get_dependencies(out)
        if isinstance(self._rhs, BaseRef):
            self._rhs._get_dependencies(out)
        return out

    def __reduce__(self):
        """Instruct pickle to not pickle the hash."""
        return type(self), (self._lhs, self._rhs)

    def __repr__(self):
        return f"({self._lhs} {self._op_str} {self._rhs})"


@cython.cclass
class UnaryOpExpr(BaseRef):
    """
    This _abstract_ class represents a unary operation on a ref.

    When overriden, the subclass must define the following:

    1. A method `_get_value` which returns the value of the expression. Tests
       have determined that this way of implementing operators (by specialising)
       classes instead of using the generic `operator` methods is ~15% faster
       in both evaluating the expressions and building the dependency graph.
    2. A cython property `_op_str` which returns the string representation of
       the operator for pretty printing.
    """
    _arg = cython.declare(object, visibility='readonly')

    def __cinit__(self, arg):
        self._arg = arg
        self._hash = hash((self.__class__, self._arg))

    def _get_value(self):
        raise NotImplementedError()

    def _get_dependencies(self, out=None):
        # The check for if `self._arg` is a ref (as for BinOpExpr) is not
        # necessary, as unless constructed manually (which should not ever
        # be done in the wild), the hypothetical literal would have been
        # evaluated to a different literal, and not this ref. Thus, for
        # performance reasons we skip the check.
        return self._arg._get_dependencies(out)

    def __reduce__(self):
        """Instruct pickle to not pickle the hash."""
        return type(self), (self._arg,)

    def __repr__(self):
        return f"({self._op_str}{self._arg})"


@cython.cclass
class AddExpr(BinOpExpr):
    _op_str = '+'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs + rhs


@cython.cclass
class SubExpr(BinOpExpr):
    _op_str = '-'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs - rhs


@cython.cclass
class MulExpr(BinOpExpr):
    _op_str = '*'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs * rhs


@cython.cclass
class MatmulExpr(BinOpExpr):
    _op_str = '@'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs @ rhs


@cython.cclass
class TruedivExpr(BinOpExpr):
    _op_str = '/'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        try:
            return lhs / rhs
        except ZeroDivisionError:
            return float('nan')


@cython.cclass
class FloordivExpr(BinOpExpr):
    _op_str = '//'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        try:
            return lhs // rhs
        except ZeroDivisionError:
            return float('nan')


@cython.cclass
class ModExpr(BinOpExpr):
    _op_str = '%'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        try:
            return lhs % rhs
        except ZeroDivisionError:
            return float('nan')


@cython.cclass
class PowExpr(BinOpExpr):
    _op_str = '**'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs ** rhs


@cython.cclass
class BitwiseAndExpr(BinOpExpr):
    _op_str = '&'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs & rhs


@cython.cclass
class BitwiseOrExpr(BinOpExpr):
    _op_str = '|'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs | rhs


@cython.cclass
class XorExpr(BinOpExpr):
    _op_str = '^'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs ^ rhs


@cython.cclass
class LtExpr(BinOpExpr):
    _op_str = '<'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs < rhs


@cython.cclass
class LeExpr(BinOpExpr):
    _op_str = '<='

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs <= rhs


@cython.cclass
class EqExpr(BinOpExpr):
    _op_str = '=='

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs == rhs


@cython.cclass
class NeExpr(BinOpExpr):
    _op_str = '!='

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs != rhs


@cython.cclass
class GeExpr(BinOpExpr):
    _op_str = '>='

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs >= rhs


@cython.cclass
class GtExpr(BinOpExpr):
    _op_str = '>'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs > rhs


@cython.cclass
class RshiftExpr(BinOpExpr):
    _op_str = '>>'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs >> rhs


@cython.cclass
class LshiftExpr(BinOpExpr):
    _op_str = '<<'

    def _get_value(self):
        lhs = BaseRef._mk_value(self._lhs)
        rhs = BaseRef._mk_value(self._rhs)
        return lhs << rhs


@cython.cclass
class NegExpr(UnaryOpExpr):
    _op_str = '-'

    def _get_value(self):
        arg = BaseRef._mk_value(self._arg)
        return -arg


@cython.cclass
class PosExpr(UnaryOpExpr):
    _op_str = '+'

    def _get_value(self):
        arg = BaseRef._mk_value(self._arg)
        return +arg


@cython.cclass
class InvertExpr(UnaryOpExpr):
    _op_str = '~'

    def _get_value(self):
        arg = BaseRef._mk_value(self._arg)
        return ~arg


@cython.cclass
class BuiltinRef(BaseRef):
    _arg = cython.declare(object, visibility='readonly')
    _op = cython.declare(object, visibility='readonly')
    _params = cython.declare(tuple, visibility='readonly')

    def __cinit__(self, arg, op, params=()):
        self._arg = arg
        self._op = op
        self._params = params
        self._hash = hash((self._op, self._arg, self._params))

    def _get_value(self):
        arg = BaseRef._mk_value(self._arg)
        return self._op(
            arg,
            *(BaseRef._mk_value(param) for param in self._params),
        )

    def _get_dependencies(self, out=None):
        arg = self._arg
        if out is None:
            out = set()
        if isinstance(arg, BaseRef):
            arg._get_dependencies(out)
        return out

    def __reduce__(self):
        """Instruct pickle to not pickle the hash."""
        return type(self), (self._op, self._args)

    def __repr__(self):
        op_symbol = OPERATOR_SYMBOLS.get(self._op, self._op.__name__)
        return f"{op_symbol}({self._arg})"


@cython.cclass
class CallRef(BaseRef):
    _func = cython.declare(object, visibility='readonly')
    _args = cython.declare(tuple, visibility='readonly')
    _kwargs = cython.declare(tuple, visibility='readonly')

    def __cinit__(self, func, args, kwargs):
        self._func = func
        self._args = args
        if isinstance(kwargs, dict):
            self._kwargs = tuple(kwargs.items())
        else:
            self._kwargs = tuple(kwargs)
        self._hash = hash((self._func, self._args, self._kwargs))

    def _get_value(self):
        func = BaseRef._mk_value(self._func)
        args = [BaseRef._mk_value(a) for a in self._args]
        kwargs = {n: BaseRef._mk_value(v) for n, v in self._kwargs}
        return func(*args, **kwargs)

    def _get_dependencies(self, out=None):
        if out is None:
            out = set()
        if isinstance(self._func, BaseRef):
            self._func._get_dependencies(out)
        for arg in self._args:
            if isinstance(arg, BaseRef):
                arg._get_dependencies(out)
        for name, arg in self._kwargs:
            if isinstance(arg, BaseRef):
                arg._get_dependencies(out)
        return out

    def __reduce__(self):
        """Instruct pickle to not pickle the hash."""
        return type(self), (self._func, self._args, self._kwargs)

    def __repr__(self):
        args = [repr(arg) for arg in self._args]
        args += [f"{k}={v}" for k, v in self._kwargs]
        args = ", ".join(args)

        if isinstance(self._func, BaseRef):
            fname = repr(self._func)
        else:
            fname = self._func.__name__

        return f"{fname}({args})"


class RefCount(dict):
    def append(self, item):
        self[item] = self.get(item, 0) + 1

    def extend(self, other):
        for kk in other:
            self.append(kk)

    def remove(self, item):
        occ = self[item]
        if occ > 1:
            self[item] = occ - 1
        else:
            del self[item]



import math

import numpy as np
import pytest

from xdeps import refs, tasks


def test_zero_divisions():
    one = refs.PosExpr(1)
    zero = refs.PosExpr(0)

    assert np.isnan((one / zero)._get_value())
    assert np.isnan((one // zero)._get_value())
    assert np.isnan((one % zero)._get_value())


def test_unary_expressions():
    arg = refs.PosExpr(7)

    assert (+arg)._get_value() == 7
    assert (-arg)._get_value() == -7
    assert (~arg)._get_value() == ~7


def test_binary_expressions():
    lhs = refs.PosExpr(7)
    rhs = refs.PosExpr(13)

    # Test standard versions
    assert (lhs + rhs)._get_value() == 7 + 13
    assert (lhs - rhs)._get_value() == 7 - 13
    assert (lhs * rhs)._get_value() == 7 * 13
    assert (lhs / rhs)._get_value() == 7 / 13
    assert (lhs // rhs)._get_value() == 7 // 13
    assert (lhs % rhs)._get_value() == 7 % 13
    assert (lhs ** rhs)._get_value() == 7 ** 13
    assert (lhs << rhs)._get_value() == 7 << 13
    assert (lhs >> rhs)._get_value() == 7 >> 13
    assert (lhs & rhs)._get_value() == 7 & 13
    assert (lhs | rhs)._get_value() == 7 | 13
    assert (lhs ^ rhs)._get_value() == 7 ^ 13
    assert (lhs > rhs)._get_value() == (7 > 13)
    assert (lhs < rhs)._get_value() == (7 < 13)
    assert (lhs >= rhs)._get_value() == (7 >= 13)
    assert (lhs <= rhs)._get_value() == (7 <= 13)

    # Test r-versions
    assert (7 + rhs)._get_value() == 7 + 13
    assert (7 - rhs)._get_value() == 7 - 13
    assert (7 * rhs)._get_value() == 7 * 13
    assert (7 / rhs)._get_value() == 7 / 13
    assert (7 // rhs)._get_value() == 7 // 13
    assert (7 % rhs)._get_value() == 7 % 13
    assert (7 ** rhs)._get_value() == 7 ** 13
    assert (7 << rhs)._get_value() == 7 << 13
    assert (7 >> rhs)._get_value() == 7 >> 13
    assert (7 & rhs)._get_value() == 7 & 13
    assert (7 | rhs)._get_value() == 7 | 13
    assert (7 ^ rhs)._get_value() == 7 ^ 13
    assert (7 > rhs)._get_value() == (7 > 13)
    assert (7 < rhs)._get_value() == (7 < 13)
    assert (7 >= rhs)._get_value() == (7 >= 13)
    assert (7 <= rhs)._get_value() == (7 <= 13)

    # Equality should check expression equivalence, instead of returning an
    # expression itself (as __eq__ implies equal hashes)
    assert ((lhs + rhs) == (lhs + refs.PosExpr(13))) is True  # equal even if ids differ
    assert ((lhs + rhs) != (lhs + 13)) is True  # unequal, as type(rhs) != type(13)
    # Actual deferred equality check:
    assert (lhs + rhs)._eq(lhs + refs.PosExpr(13))._get_value() is True
    assert (lhs + rhs)._eq(lhs + 13)._get_value() is True
    assert (lhs + rhs)._eq(lhs)._get_value() is False

    assert (lhs + rhs)._neq(lhs + refs.PosExpr(13))._get_value() is False
    assert (lhs + rhs)._neq(lhs + 13)._get_value() is False
    assert (lhs + rhs)._neq(lhs)._get_value() is True


def test_matmul_expression():
    class DummyHashableMatrix:
        def __init__(self, value):
            self._matrix = np.array(value)

        def __matmul__(self, other):
            return DummyHashableMatrix(self._matrix @ other._matrix)

        def __pos__(self):
            return self

        def __hash__(self):
            return hash(self._matrix.tobytes())

        def __eq__(self, other):
            return np.all(self._matrix == other._matrix)

    matlhs = refs.PosExpr(DummyHashableMatrix([[1, 2], [3, 4]]))
    matrhs = refs.PosExpr(DummyHashableMatrix([[5], [6]]))
    assert (matlhs @ matrhs)._get_value() == DummyHashableMatrix([[17], [39]])


def test_matmul_expression_ref():
    container = {
        'a': np.array([[1, 2], [3, 4]]),
        'b': np.array([[5], [6]]),
    }

    manager = tasks.Manager()
    ref = manager.ref(container, 'ref')
    ref['c'] = ref['a'] @ ref['b']

    assert np.all(container['c'] == np.array([[17], [39]]))


def test_builtin_expression():
    lhs = refs.PosExpr(7)
    rhs = refs.PosExpr(13)
    fp = refs.PosExpr(-7.13)

    assert divmod(lhs, rhs)._get_value() == divmod(7, 13)
    assert round(fp)._get_value() == round(-7.13)
    assert math.trunc(fp)._get_value() == math.trunc(-7.13)
    assert math.floor(fp)._get_value() == math.floor(-7.13)
    assert math.ceil(fp)._get_value() == math.ceil(-7.13)
    assert abs(fp)._get_value() == abs(-7.13)
    # These will not work, as Python verifies the type of the result:
    # assert complex(lhs, rhs)._get_value() == complex(7, 13)
    # assert int(fp)._get_value() == int(-7.13)
    # assert float(lhs)._get_value() == float(7)


def test_handle_inexistent_values():
    manager = tasks.Manager()
    container = {
        'dict': {},
        'object': object(),
    }
    ref = manager.ref(container, 'ref')

    with pytest.raises(AttributeError):
        ref['object'].not_there._get_value()

    with pytest.raises(BaseException):
        # See the comment in BaseRef._value for why it's not an AttributeError
        _ = ref['object'].not_there_either._value

    with pytest.raises(KeyError):
        ref['dict']['not there']._get_value()

    with pytest.raises(KeyError):
        _ = ref['dict']['not there either']._value


@pytest.mark.xfail(reason='Undefined behaviour')
def test_ref_inplace_ops():
    manager = tasks.Manager()
    container = np.array([1, 2])
    ref = manager.ref(container, 'ref')

    assert np.all(ref._get_value() == [1, 2])
    ref += 3
    assert np.all(ref._get_value() == [4, 5])

    # The below does not work, as it is not clear how to implement this in a way
    # that is universal. Is it even desired?
    assert np.all(container == [4, 5])


def test_itemref():
    manager = tasks.Manager()
    container = {'a': 1, 'b': 2}
    ref = manager.ref(container, 'ref')

    item_ref_a = ref['a']
    assert item_ref_a._get_value() == 1

    item_ref_b = ref['b']
    assert item_ref_b._get_value() == 2
    item_ref_b._set_value(3)
    assert item_ref_b._get_value() == 3


def test_attrref():
    class DummyClass:
        def __init__(self, a, b):
            self.a = a
            self.b = b

    manager = tasks.Manager()
    container = DummyClass(1, 2)
    ref = manager.ref(container, 'ref')

    attr_ref_a = ref.a
    assert attr_ref_a._get_value() == 1

    attr_ref_b = ref.b
    assert attr_ref_b._get_value() == 2
    attr_ref_b._set_value(3)
    assert attr_ref_b._get_value() == 3


def test_call_ref():
    def func(arg1, arg2, *, kwarg1, kwarg2):
        return arg1, arg2, kwarg1, kwarg2

    manager = tasks.Manager()
    container = {'arg1': 1, 'kwarg1': 3}
    data_ref = manager.ref(container, 'container')

    manager = tasks.Manager()
    ref = manager.ref(func, 'func')

    call_ref = ref(data_ref['arg1'], 4, kwarg2=2, kwarg1=data_ref['kwarg1'])
    assert call_ref._get_value() == (1, 4, 3, 2)

    assert call_ref._get_dependencies() == {data_ref['arg1'], data_ref['kwarg1']}


def test_objectattrref():
    container = {'a': 3}
    manager = tasks.Manager()
    ref = manager.refattr(container, 'ref')

    assert ref.a._get_value() == 3

    ref.b = 42
    assert ref.b._get_value() == 42
    assert container['b'] == 42


def test_refcount():
    ref_count = refs.RefCount()
    ref_count.extend(['a', 'b'])
    ref_count.append('b')

    assert ref_count['a'] == 1
    assert ref_count['b'] == 2

    ref_count.append('b')
    assert ref_count['a'] == 1
    assert ref_count['b'] == 3

    ref_count.remove('a')
    assert 'a' not in ref_count
    assert ref_count['b'] == 3

    ref_count.remove('b')
    assert 'a' not in ref_count
    assert ref_count['b'] == 2


def test_cythonized():
    assert refs.is_cythonized()



import os
import pathlib
import re
from typing import Collection
from collections import namedtuple

import numpy as np

gblmath = {"np": np}
for k, fu in np.__dict__.items():
    if type(fu) is np.ufunc:
        gblmath[k] = fu


def is_iterable(obj):
    return hasattr(obj, "__iter__") and not isinstance(obj, str)


def eval_col(col, view):
    try:
        cc = eval(col, gblmath, view)
    except NameError:
        raise KeyError(
            f"Column `{col}` could not be found or " "is not a valid expression"
        )
    return cc


def _to_str(arr, digits, fixed="g", max_len=None):
    """Covert an array to an array of string representations.

    Parameters
    ----------
    arr : array-like
        Array to convert.
    digits : int
        Number of digits to use for floats.
    fixed : str
        If 'g', use general format (total number of digits is `digits`, if
        necessary use exponent notation). If 'f', use fixed point notation
        (exactly `digits` digits after the decimal point).
    max_len : int or None
        If not None, truncate strings to this length.
    """
    if len(arr.shape) > 1:
        # If array of arrays, just show the shape
        out = np.repeat(f"<array of shape {arr.shape[1:]}>", arr.shape[0])
    elif arr.dtype.kind in "SU":
        # Keep strings
        out = arr
    elif arr.dtype.kind in "iu":
        # Print out integers in full
        out = np.char.mod("%d", arr)
    elif arr.dtype.kind in "f":
        # Format floats.
        # First generate a format string like "%.<digits>g", then use it on
        # each element.
        add = 3 if fixed == "f" else 7
        fmt = "%%%d.%d%s" % (digits + add, digits, fixed)
        out = np.char.mod(fmt, arr)
    elif arr.dtype.kind == "O" and isinstance(arr[0], Collection):
        # If array of collections (array with dtype=object) or list, give shape
        lengths = []
        for entry in arr:
            if isinstance(entry, str):
                lengths.append(entry)
            elif isinstance(entry, np.ndarray):
                lengths.append(f"<array of shape {entry.shape}>")
            elif isinstance(entry, Collection):
                lengths.append(f"<collection of length {len(entry)}>")
            else:
                lengths.append(str(entry))
        out = np.array(lengths)
    else:
        # Any other flat array: stringify.
        out = arr.astype("U")

    if max_len is not None:
        old_out = out.copy()
        out = []
        for ii in range(len(old_out)):
            out.append(
                str(
                    old_out[ii][:max_len] + " ..."
                    if len(old_out[ii]) > max_len
                    else old_out[ii]
                )
            )
        out = np.array(out)
    return out


class _View:
    def __init__(self, data, index, nrows):
        self.data = data
        self.index = index
        self.nrows = nrows

    def __getitem__(self, k):
        return self.data[k][self.index]

    def __len__(self):
        return len(self.data)

    def get(self, k, default=None):
        if k == "__tracebackhide__":  # to avoid issues in ipython
            return None
        return self.data.get(k, default)[self.index]

    def __repr__(self):
        return f"<View {sum(self.index)} rows>"

    def __contains__(self, k):
        return k in self.data

    def eval(self, col):
        try:
            cc = eval(col, gblmath, self)
        except NameError:
            raise KeyError(
                f"Column `{col}` could not be found or " "is not a valid expression"
            )
        return cc

    def get_indices(self):
        if hasattr(self.data, "get_indices"):
            return self.data.get_indices()[self.index]
        else:
            return np.arange(self.nrows)[self.index]


class Table:
    """Table class managing list of columns and scalars values. Columns are numpy
    arrays with same lengths. A special column can be assigned as string index.

    The table can be accessed by columns, rows or by a combination of both. The table
    can be transposed and shown in a human readable format.

    Main extraction API:

    - `table[col]` : get a column (numpy array) or a scalar value
    - `table[col,row]` : get a value at a specific row  which can be:
        - integer
        - string in the form 'name::count<<offset'
        - tuple (name, count, offset) or (name, count)
    - `table.rows[<sel1>,<sel2>,...]` : get a view of the table by rows. <sel> can be:
        - a numerical index
        - a regular expression (beware of special characters!)
        - a slice (start:stop:step) or a range (low:high:column)
        - a combination of above equivalent to table.rows[<sel1>].rows[<sel2>].rows[...]
    - `table.cols[<col1>,<col2>,...]`: get a view of the table by columns, where column can be also valid expression of column names.

    Other methods and properties:
    - `table.show()` : show the table in a human readable format
    - `table._t` : transpose the table
    - `table._df` : convert the table to a pandas DataFrame

    Row view API:

    - `table.rows.at(index, as_dict=False)` : get a row as a namedtuple or dictionary
    - `table.rows.head(n=5)` : get the first n rows
    - `table.rows.tail(n=5)` : get the last n rows
    - `table.rows.reverse()` : reverse the order of the rows
    - `table.rows.transpose()` : transpose the table
    - `table.rows.indices[...]` : get indices from rows selectors
    - `table.rows.mask[...]` : get boolena mask from rows selectors
    - `table.rows.is_repeated(row)` : check if a row is repeated

    Column view API:
    - `table.cols.names` : list of column names
    - `table.cols.keys()` : iterator over column names
    - `table.cols.values()` : iterator over column values
    - `table.cols.items()` : iterator over column name and values
    - `table.cols.transpose()` : transpose the table


    Public low-level API:

    - `table._col_names` : list of column names
    - `table._index` : column name used as index
    - `table._sep_count` : separator for count in index column
    - `table._sep_previous` : separator for previous index column
    - `table._sep_next` : separator for next index column
    - `table._regex_flags` : flags for regex matching
    - `table._select(rows, cols)` : select a subtable by rows and columns
    - `table._select_rows(rows)` : select a subtable by rows as accepted by numpy array
    - `table._select_cols(cols)` : select a subtable by iterable of column names
    """

    def __init__(
        self,
        data,
        col_names=None,
        index="name",
        sep_count="::",
        sep_previous="<<",
        sep_next=">>",
        cast_strings=True,
        regex_flags=re.IGNORECASE,
        verify=True,
    ):
        if verify:
            _data = data.copy()
            _col_names = list(data.keys() if col_names is None else col_names)
            for kk in _col_names:
                vv = data[kk]
                if not hasattr(vv, "dtype"):
                    raise ValueError(f"Column `{kk}` is not a numpy array")
                else:
                    if cast_strings and vv.dtype.kind in "SU":
                        vv = np.array(vv, dtype=object)
                _data[kk] = vv

            nrows = set(len(_data[cc]) for cc in _col_names)
            if len(nrows) > 1:
                for cc in _col_names:
                    print(f"Length {cc!r} = {len(_data[cc])}")
                raise ValueError("Columns have different lengths")

            if index is not None and index not in _col_names:
                raise ValueError(f"Index column `{index}` not found in columns")
        else:
            _data = data
            _col_names = col_names

        # special init due to setattr redefinition
        init = {
            #            "header": header,
            "cols": _ColView(self),
            "rows": _RowView(self),
            "_data": _data,
            "_col_names": _col_names,
            "_index": index,
            "_index_cache": None,
            "_count_cache": None,
            "_sep_count": sep_count,
            "_sep_previous": sep_previous,
            "_sep_next": sep_next,
            "_regex_flags": regex_flags,
        }

        for kk, vv in init.items():
            object.__setattr__(self, kk, vv)

    @property
    def mask(self):
        raise DeprecationWarning(
            "mask is deprecated, use table.rows.mask or consider using table.rows.indices"
        )

    def _get_row_where_col(self, col, row, count=0, offset=0):
        # generally slower than _get_row_col_fast
        return np.where(col == row)[0][count] + offset

    def _make_cache(self):
        col = self._data[self._index]
        dct = {}
        count = {}
        newnames = np.zeros(len(col), dtype=object)
        for ii, nn in enumerate(col):
            cc = count.get(nn, -1) + 1
            dct[(nn, cc)] = ii
            count[nn] = cc
            newnames[ii] = f"{nn}{self._sep_count}{cc}"
        for nn, cc in count.items():
            count[nn] = cc + 1
            if cc == 0:
                newnames[dct[(nn, 0)]] = nn
        return dct, count, newnames

    def _get_cache(self):
        if self._index_cache is None:
            _index_cache, _count_cache, _names_cache = self._make_cache()
            object.__setattr__(self, "_index_cache", _index_cache)
            object.__setattr__(self, "_count_cache", _count_cache)
            object.__setattr__(self, "_names_cache", _names_cache)
        return self._index_cache, self._count_cache

    def _get_row_cache(self, row, count, offset):
        """Get the index of a row by name and repetition."""
        cache, count_dict = self._get_cache()
        if count is None:
            count = 0
        if count < 0:
            count += count_dict[row]
        idx = cache.get((row, count))
        return idx + offset if idx is not None else None

    def _get_row_cache_raise(self, row, count=0, offset=0):
        idx = self._get_row_cache(row, count, offset)
        if idx is None:
            raise KeyError(f"Cannot find '{row}' in column '{self._index}'")
        return idx

    def _split_name_count_offset(self, name):
        """Split a name::count<<offset into name, count, and offset."""

        # Initialize variables
        count = None
        offset = 0

        # Check for offset with self._sep_previous and self._sep_next
        if self._sep_previous in name:
            name, offset_str = name.split(self._sep_previous, 1)
            offset -= int(offset_str)
        elif self._sep_next in name:
            name, offset_str = name.split(self._sep_next, 1)
            offset += int(offset_str)

        # Check for count with self._sep_count
        if self._sep_count in name:
            name, count_str = name.split(self._sep_count, 1)
            count = int(count_str)

        return name, count, offset

    def _split_name_count_using_re(self, name):
        ssplit = re.compile(r"^([^:<>]*)(::([+-]?\d+))?(<<([+-]?\d+))?(<<([+-]?\d+))?")
        gg = ssplit.match(name).groups()
        name, _, count, _, prev, _, next = gg
        count = None if count is None else int(count)
        offset = 0 if prev is None else -int(prev)
        offset += 0 if next is None else int(next)
        return name, count, offset

    def _get_regexp_indices(self, regexp):
        """Get indices using string selector on the index column.

        Examples:
           "mq.*::1<<1" -> all elements preceding second matches of mq.*
        """
        name, count, offset = self._split_name_count_offset(regexp)
        if count is not None:
            tryidx=self._get_row_cache(name, count, offset)
            if tryidx is not None:
                return np.array([tryidx], dtype=int)
        regexpr = re.compile(name, flags=self._regex_flags)
        iilst = []
        nnlst = set()
        for ii, nn in enumerate(self._data[self._index]):
            if regexpr.fullmatch(nn):
                iilst.append(ii)
                nnlst.add(nn)
        if count is not None:
            iilst=[]
            for nn in nnlst:
                idx=self._get_row_cache(nn, count,0)
                if idx is not None:
                    iilst.append(idx)
        return np.array(iilst, dtype=int) + offset

    def _get_col_regexp_indices(self, regexp, col):
        """Get indices using string selector on a column."""
        regexpr = re.compile(regexp, flags=self._regex_flags)
        lst = []
        for ii, nn in enumerate(self._data[col]):
            if regexpr.fullmatch(nn):
                lst.append(ii)
        return np.array(lst, dtype=int)

    def _get_row_index(self, row):
        """
        get a row index from:
        - integer
        - name::count<<offset, name::count>>offset
        - (name, count, offset) or (name, count)
        """
        if isinstance(row, int):
            return row
        elif isinstance(row, str):
            row, count, offset = self._split_name_count_offset(row)
            return self._get_row_cache_raise(row, count, offset)
        elif isinstance(row, tuple):
            return self._get_row_cache_raise(*row)
        else:
            raise ValueError(f"Invalid row {row}")

    def _get_row_indices(self, row):
        """
        get the row indices from:
        - integer
        - regexp::count<<offset, name::count>>offset
        - (name, count, offset) or (name, count)
        - slice
        - list of the above
        """
        if isinstance(row, slice):
            ia = row.start
            ib = row.stop
            ic = row.step
            if isinstance(ia, str) or isinstance(ib, str):  # name matching
                if ic is None or ic == self._index:
                    if ia is not None:
                        ia = self._get_row_index(ia)
                    if ib is not None:
                        ib = self._get_row_index(ib) + 1
                else:
                    if ia is not None:
                        ia = self._get_row_where_col(self._data[ic], ia)
                    if ib is not None:
                        ib = self._get_row_where_col(self._data[ic], ib) + 1
                return slice(ia, ib)
            elif isinstance(ic, str):  # range matching
                col = self._data[ic]
                if ia is None and ib is None:
                    return slice(None)
                elif ia is not None and ib is None:
                    return np.where(col <= ib)[0]
                elif ib is not None and ia is None:
                    return np.where(col >= ia)[0]
                else:
                    return np.where((col >= ia) & (col <= ib))[0]
            else:  # plain slice
                return row
        elif isinstance(row, str):
            return self._get_regexp_indices(row)
        elif is_iterable(row):  # could be a mask
            if len(row) == 0:
                return np.array([], dtype=int)
            row = np.array(row)
            if row.dtype is np.dtype("bool"):
                return np.where(row)[0]
            elif row.dtype.kind in "SU":
                return np.fromiter(
                    map(self._get_row_index, row), dtype=int, count=len(row)
                )
            elif row.dtype.kind in "iu":
                return row
            elif row.dtype.kind == "O":
                out = []
                for rr in row:
                    if isinstance(rr, str):
                        out.append(self._get_row_index(rr))
                    elif isinstance(rr, int):
                        out.append(rr)
                    else:
                        raise ValueError(f"Invalid row {rr}")
                return np.array(out, dtype=int)
            else:
                raise ValueError(f"Invalid row selector {row}")
        else:
            return [self._get_row_index(row)]

    def _get_name_mask(self, name, col):
        """Get mask using string selector on a column."""
        mask = np.zeros(self._nrows, dtype=bool)
        indices = self._get_row_indices(name)
        mask[indices] = True
        return mask

    def _select(self, rows, cols):
        view = self._data
        # create row view
        if is_iterable(rows):
            for row in rows:
                if row is not None:
                    view = _View(view, self._get_row_indices(row), len(self))
        else:
            if rows is not None:
                view = _View(view, self._get_row_indices(rows), len(self))
        # select columms
        if cols is None or cols == slice(None, None, None):
            col_list = self._col_names
        elif isinstance(cols, str):
            col_list = cols.split()
        elif is_iterable(cols):
            col_list = list(cols)
        else:
            raise ValueError(f"Invalid column: {cols}")

        if self._index not in col_list:
            col_list.insert(0, self._index)
        data = {}
        for cc in col_list:
            data[cc] = eval(cc, gblmath, view)
        for kk in self.keys(exclude_columns=True):
            data[kk] = self._data[kk]
        return self.__class__(
            data,
            col_names=col_list,
            index=self._index,
            sep_count=self._sep_count,
            sep_previous=self._sep_previous,
            sep_next=self._sep_next,
            regex_flags=self._regex_flags,
            verify=False,
        )

    def _select_rows(self, rows):
        """Select a subtable by rows as accepted by numpy array."""
        data = {}
        for cc in self._col_names:
            data[cc] = self._data[cc][rows]
        for kk in self.keys(exclude_columns=True):
            data[kk] = self._data[kk]
        return self.__class__(
            data,
            col_names=self._col_names,
            index=self._index,
            sep_count=self._sep_count,
            sep_previous=self._sep_previous,
            sep_next=self._sep_next,
            regex_flags=self._regex_flags,
            verify=False,
        )

    def _select_cols(self, cols):
        """Select a subtable by iterable of column names."""
        data = {}
        for cc in cols:
            data[cc] = self._data[cc]
        for kk in self.keys(exclude_columns=True):
            data[kk] = self._data[kk]
        if self._index not in cols:
            cols.insert(0, self._index)
            data[self._index] = self._data[self._index]
        return self.__class__(
            data,
            col_names=cols,
            index=self._index,
            sep_count=self._sep_count,
            sep_previous=self._sep_previous,
            sep_next=self._sep_next,
            regex_flags=self._regex_flags,
            verify=False,
        )

    def __getitem__(self, args):
        """Extract data from the table.

        If one argument is given, it can be a column name or a valid expression
        of column names. If multiple arguments are given, the first argument is
        the column name or expression and the following arguments are row
        selectors.
        """
        if isinstance(args, str):
            try:
                return self._data[args]
            except KeyError:
                return eval(args, gblmath, self._data)
        if type(args) is tuple:  # multiple args
            if len(args) == 0:
                col = None
                row = None
            elif len(args) == 1:
                col = args[0]
                row = None
            elif len(args) == 2:
                col = args[0]
                row = args[1]
                try:
                    col = self._data[col]
                except KeyError:
                    col = eval(col, gblmath, self._data)
                if isinstance(row, str):
                    cache, count = self._get_cache()
                    idx = cache.get((row, 0))
                    if idx is None:
                        name, count, offset = self._split_name_count_offset(row)
                        idx = self._get_row_cache_raise(name, count, offset)
                elif isinstance(row, tuple):
                    cache, count = self._get_cache()
                    idx = cache.get(row)
                    if idx is None:
                        idx = self._get_row_cache_raise(*row)
                elif isinstance(row, slice):
                    idx = self._get_row_indices(row)
                elif isinstance(row, list):
                    idx = self._get_row_indices(row)
                else:
                    idx = row
                return col[idx]
            else:
                raise ValueError(
                    f"Too many arguments {args} for <Table id={id(self)}>."
                )
        raise ValueError(f"Invalid arguments {args} for <Table id={id(self)}>.")

    def show(
        self,
        rows=None,
        cols=None,
        maxrows=None,
        maxwidth="auto",
        max_col_width=None,
        output=None,
        digits=6,
        fixed="g",
        header=True,
    ):
        """Show the table in a human readable format.

        Parameters:
        - rows : string, slice or list or None
            Rows to show. If None, show all rows. See table.rows for more info.
        - cols : string, list or None
            Columns to show. If None, show all columns. See table.cols for more info.
        - maxrows : int or None
            Maximum number of rows to show. If None, show all rows.
        - maxwidth : int, "auto" or "full"
            Maximum width of the output. If "auto", use the terminal width.
            If "full", use the full width.
        - output : None, str or file-like object
            If None, print the output. If str, return the output as a string.
            If file-like object, write the output to the file.
        - digits : int
            Number of digits to use for floats.
        - fixed : str
            If 'g', use general format (total number of digits is `digits`, if
            necessary use exponent notation). If 'f', use fixed point notation
            (exactly `digits` digits after the decimal point).
        - header : bool
            If True, show the header.
        - max_col_width : int or None
            Maximum width of a column. If None, use the terminal width.

        """
        if len(self) == 0:
            return ""
        unique = self._make_cache()[2]
        if rows is None and cols is None:
            view = self
        else:
            view = self._select(rows, cols)
            indices = view._get_row_indices(rows)
            unique = unique[indices]

        col_list = view._col_names

        if self._index is not None:
            # index always first
            if self._index in col_list:
                col_list.remove(self._index)
            col_list.insert(0, self._index)

        cut = -1
        viewrows = len(view)
        if maxrows is not None and output is None and viewrows > maxrows:
            cut = maxrows // 2

        if maxwidth == "auto":
            try:
                maxwidth = os.get_terminal_size().columns - 5
                if maxwidth < 10 or maxwidth > 10000:
                    raise ValueError("Terminal width too big or too small.")
            except (OSError, ValueError):
                maxwidth = 100
        elif maxwidth == "full" or maxwidth is None:
            maxwidth = 100000000

        data = []
        width = 0
        fmt = []
        header_line = []
        for cc in col_list:
            if cc == self._index:
                coldata = unique
            elif cc in view:
                coldata = view[cc]
            else:
                coldata = eval(cc, gblmath, view)
            if cut > 0:
                coldata = np.r_[coldata[:cut], coldata[cut:]]
            coltype = coldata.dtype.kind
            col = _to_str(coldata, digits, fixed, max_len=max_col_width)
            colwidth = int(col.dtype.str[2:])
            if len(cc) > colwidth:
                colwidth = len(cc)
            width += colwidth + 1
            if width < maxwidth:
                if coltype in "SUO":
                    fmt.append("%%-%ds" % (colwidth))
                else:
                    fmt.append("%%%ds" % colwidth)
                header_line.append(fmt[-1] % str(cc))
                data.append(col)
            else:
                header_line.append("...")
                break

        result = []
        if header:
            result.append(" ".join(header_line))
        for ii in range(len(col)):
            row = " ".join([ff % col[ii] for ff, col in zip(fmt, data)])
            result.append(row)
            if ii == cut:
                result.append("...")
        result = "\n".join(result)
        if output is None:
            print(result)
        elif output is str:
            return result
        elif hasattr(output, "write"):
            output.write(result)
        else:
            output = pathlib.Path(output)
            with open(output, "w") as fh:
                fh.write(result)

    def __repr__(self):
        n = len(self)
        c = len(self._col_names)
        ns = "s" if n != 1 else ""
        cs = "s" if c != 1 else ""
        out = [f"{self.__class__.__name__}: {n} row{ns}, {c} col{cs}"]
        if n ==0:
            pass
        elif n < 30:
            out.append(self.show(output=str, maxwidth="auto"))
        else:
            out.append(self.show(rows=slice(0, 10), output=str, maxwidth="auto"))
            out.append("...")
            out.append(
                self.show(
                    rows=slice(-10, None), header=False, output=str, maxwidth="auto"
                )
            )
            # out.append("Use `table.show()` to see the full table.")
        return "\n".join(out)

    @classmethod
    def from_pandas(cls, df, index=None, lowercase=False):
        if index is None:
            index = df.index.name
            if index is None and "NAME" in df.columns:
                index = "NAME"
        if lowercase:
            df.columns = df.columns.str.lower()
            index = index.lower()
            for cc in df.columns:
                if df[cc].dtype.kind in "SUO":
                    df[cc] = df[cc].str.lower()
        col_names = list(df.columns)
        data = {cc: df[cc].values for cc in col_names}
        if hasattr(df, "headers"):
            for cc, dd in df.headers.items():
                cc = cc.lower() if lowercase else cc
                if cc in data:
                    cc = cc + "_header"
                if lowercase and isinstance(dd, str):
                    dd = dd.lower()
                data[cc] = dd
        return cls(data, col_names=col_names, index=index)

    @classmethod
    def from_csv(cls, filename, index=None, **kwargs):
        import pandas as pd

        df = pd.read_csv(filename, **kwargs)
        if index is None and "NAME" in df.columns:
            index = "NAME"

        return cls.from_pandas(df, index=index)

    @classmethod
    def from_rows(cls, rows, col_names=None, index=None):
        if hasattr(rows[0], "_asdict"):  # namedtuple
            if col_names is None:
                col_names = list(rows[0]._fields)
            data = {cc: np.array([getattr(rr, cc) for rr in rows]) for cc in col_names}
        else:
            if col_names is None:
                col_names = list(rows[0].keys())
            data = {cc: np.array([rr[cc] for rr in rows]) for cc in col_names}
        return cls(data, col_names=col_names, index=index)

    @classmethod
    def from_tfs(cls, filename, index=None, lowercase=True):
        from tfs import read_tfs

        df = read_tfs(filename)
        return cls.from_pandas(df, index=index, lowercase=lowercase)

    @property
    def _t(self):
        """Transpose the table."""
        data = {"columns": np.array(self._col_names)}
        for nn in range(len(self)):
            data[f"row{nn}"] = np.array([str(self[cc][nn]) for cc in self._col_names])
        return Table(data, index="columns", col_names=list(data.keys()))

    @property
    def _df(self):
        return self.to_pandas()

    def __contains__(self, key):
        return self._data.__contains__(key)

    def __delitem__(self, key):
        if key in self._col_names:
            self._col_names.remove(key)
        del self._data[key]

    def __dir__(self):
        return super().__dir__() + list(self._data.keys())

    def __iter__(self):
        return self._data.__iter__()

    def __getstate__(self):
        return self.__dict__

    def __setstate__(self, state):
        self.__dict__.update(state)

    def __getattr__(self, key):
        if key in self._data:
            return self._data[key]
        else:
            raise AttributeError(f"Cannot find `{key}` in table")

    def __len__(self):
        k = self._col_names[0]
        return len(self._data[k])

    def __setitem__(self, key, val):
        if key == self._index or key=="_sep_count":
            object.__setattr__(self, "_index_cache", None)
            object.__setattr__(self, "_count_cache", None)
            object.__setattr__(self, "_name_cache", None)
        if key in self.__dict__:
            object.__setattr__(self, key, val)
        elif key in self._col_names:
            self._data[key][:] = val
        else:
            self._data[key] = val
            if hasattr(val, "__iter__") and len(val) == len(self):
                self._col_names.append(key)

    __setattr__ = __setitem__

    def __neg__(self):
        return self.rows.reverse()

    def _concatenate_table(self, table):
        """Concatenate a table to the table."""
        for col in table._col_names:
            self._data[col] = np.concatenate([self._data[col], table._data[col]])
        return self

    def __add__(self, other):
        if isinstance(other, Table):
            res = self._copy()
            return res._concatenate_table(other)
        else:
            raise ValueError("Can only concatenate Table with Table")

    def __mul__(self, num):
        res = self._copy()
        for col in res._col_names:
            res._data[col] = np.concatenate([res._data[col]] * num)
        return res

    def __floordiv__(self, name):
        return self._get_row_index(name)

    def _copy(self):
        return self.__class__(
            self._data.copy(),
            col_names=self._col_names,
            index=self._index,
            sep_count=self._sep_count,
            sep_previous=self._sep_previous,
            sep_next=self._sep_next,
        )

    def items(self):
        return self._data.items()

    def keys(self, exclude_columns=False):
        if exclude_columns:
            return set(self._data) - set(self._col_names)
        return self._data.keys()

    def values(self):
        return self._data.values()

    def pop(self, key):
        res = self._data.pop(key)
        if key in self._col_names:
            self._col_names.remove(key)
        return res

    def _update(self, data):
        """Update the table with new data."""
        if hasattr(self, "_data"):
            data = self._data
        for name, value in data.items():
            if hasattr(value, "__len__") and len(value) == len(self):
                if name not in self._data:
                    self._col_names.append(name)
            self._data[name] = value

    def _append_row(self, row):
        """Append a row to the table."""
        for col in self._col_names:
            self._data[col] = np.r_[self._data[col], [row[col]]]

    def to_pandas(self, index=None, columns=None):
        if columns is None:
            columns = self._col_names

        import pandas as pd

        df = pd.DataFrame(self._data, columns=self._col_names)
        if index is not None:
            df.set_index(index, inplace=True)
        return df


class Indices:
    """Class returing indices of the table.
    t.indices[1] -> row
    t.indices['a'] -> pattern
    r.indices[10:20]-> range
    r.indices['a':'b'] -> name range
    r.indices['a':'b':'myname'] -> name range with 'myname' column
    r.indices[-2:2:'x'] -> name value range with 'x' column
    r.indices[-2:2:'x',...] -> & combinations
    """

    def __init__(self, table):
        self.table = table

    def __getitem__(self, rows):
        if isinstance(rows, tuple):  # multiple arguments
            return self.table.rows._make_view(*rows).get_indices()
        else:
            return self.table._get_row_indices(rows)


class Mask:
    def __init__(self, table):
        self.table = table

    def __getitem__(self, rows):
        indices = self.table.rows.indices[rows]
        mask = np.zeros(len(self.table), dtype=bool)
        mask[indices] = True
        return mask


class _RowView:
    def __init__(self, table):
        self.table = table
        self.mask = Mask(table)
        self.indices = Indices(table)

    def _make_view(self, *rows):
        view = self.table
        for row in rows:
            view = _View(view, self.table._get_row_indices(row), len(self.table))
        return view

    def __getitem__(self, rows):
        if isinstance(rows, tuple):  # multiple arguments
            indices = self.indices[rows]
        else:
            indices = self.table._get_row_indices(rows)
        return self.table._select_rows(indices)

    def __iter__(self):
        res_type = namedtuple("Row", self.table._col_names)
        for ii in range(len(self.table)):
            yield res_type(*[self.table[cc, ii] for cc in self.table._col_names])

    def at(self, index, as_dict=False):
        if as_dict:
            return {cc: self.table[cc, index] for cc in self.table._col_names}
        else:
            res_type = namedtuple("Row", self.table._col_names)
            return res_type(*[self.table[cc, index] for cc in self.table._col_names])

    def transpose(self):
        return self.table._t

    def __repr__(self):
        return f"RowView: {len(self.table)} rows, {len(self.table.cols)} cols"

    def __len__(self):
        return len(self.table)

    def head(self, n=5):
        return self[:n]

    def tail(self, n=5):
        return self[-n:]

    def reverse(self):
        return self[::-1]

    def is_repeated(self, row):
        _, count_dict = self.table._get_cache()
        return count_dict.get(row, 0) > 1

    def get_index_fast(self, name, count=0, offset=0):
        """Get the index of a row by name and repetition and offset."""
        index_cache, _ = self.table._get_cache()
        return index_cache[(name, count)] + offset

    def get_index(self, row):
        """Get the index of a row by:
        - a string in the form 'name::count<<offset' or 'name::count>>offset'
        - a tuple (name, count, offset)
        """
        return self.table._get_row_index(row)

    def get_regexp_indices(self, regexp):
        """Get indices using string selector on the index column.
        """
        return self.table._get_regexp_indices(regexp)
 
    def get_col_regexp_indices(self, regexp, col):
        """Get indices using regular expression on the index column.
        """
        return self.table._get_col_regexp_indices(regexp, col)


class _ColView:
    def __init__(self, table):
        self.table = table

    def __getitem__(self, cols):
        if cols is None or cols == slice(None, None, None):
            col_list = self._col_names
        elif isinstance(cols, str):
            col_list = cols.split()
        elif is_iterable(cols):
            col_list = list(cols)
        else:
            raise ValueError(f"Invalid column: {cols}")

        if self.table._index is not None and self.table._index not in col_list:
            col_list.insert(0, self.table._index)
        return self.table._select_cols(col_list)

    @property
    def names(self):
        return self.table._col_names

    def __repr__(self):
        return "ColView: " + " ".join(self.table._col_names)

    def keys(self):
        return iter(self.table._col_names)

    def values(self):
        return (self.table._data[cc] for cc in self.table._col_names)

    def items(self):
        return ((cc, self.table._data[cc]) for cc in self.table._col_names)

    def __contains__(self, key):
        return key in self.table._col_names

    def __iter__(self):
        return iter(self.table._col_names)

    def __len__(self):
        return len(self.table._col_names)

    def transpose(self):
        return self.table._t

    def get_index_unique(self):
        """Get index column with unique rows."""
        _, _, names = self.table._make_cache()
        return names



# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from Cython.Build import cythonize
from setuptools import setup, find_packages
from pathlib import Path

version_file = Path(__file__).parent / 'xdeps/_version.py'
dd = {}
with open(version_file.absolute(), 'r') as fp:
    exec(fp.read(), dd)
__version__ = dd['__version__']

setup(
    name='xdeps',
    version=__version__,
    description='Data dependency manager',
    long_description=("Data dependency manager\n"
                     "\nThis package is part of the Xsuite collection."),
    author='Riccardo De Maria',
    author_email='riccardo.de.maria@cern.ch',
    packages=find_packages(),
    install_requires=['lark', 'numpy', 'scipy', 'cython'],
    url='https://xsuite.readthedocs.io/',
    license='Apache 2.0',
    download_url="https://pypi.python.org/pypi/xdeps",
    project_urls={
        "Bug Tracker": "https://github.com/xsuite/xsuite/issues",
        "Documentation": 'https://xsuite.readthedocs.io/',
        "Source Code": "https://github.com/xsuite/xdeps",
    },
    extras_require={
        'tests': ['pytest'],
    },
    ext_modules=cythonize("xdeps/refs.py"),
    extra_compile_args=["-O3"],
)



# copyright ############################### #
# This file is part of the Xdeps Package.   #
# Copyright (c) CERN, 2021.                 #
# ######################################### #

from collections import defaultdict
import math
import numpy as np

from lark import Lark, Transformer, v_args
from .tasks import Manager
from .utils import AttrDict
from .table import Table

calc_grammar = """
    ?start: sum
        | NAME "=" sum      -> assign_var

    ?sum: product
        | sum "+" product   -> add
        | sum "-" product   -> sub

    ?product: power
        | product "*" power  -> mul
        | product "/" power  -> div

    ?power: atom
        | power "^" atom    -> pow

    ?atom: NUMBER           -> number
         | "-" atom         -> neg
         | "+" atom         -> pos
         | NAME             -> var
         | NAME "->" NAME   -> getitem
         | NAME "(" sum ("," sum)* ")" -> call
         | "(" sum ")"

    NAME: /[a-z_\\.][a-z0-9_\\.%]*/
    %import common.NUMBER
    %import common.WS_INLINE
    %ignore WS_INLINE
"""


@v_args(inline=True)
class MadxEval(Transformer):
    from operator import add, sub, mul, truediv as div
    from operator import neg, pos, pow

    number = float

    def __init__(self, variables, functions, elements, get="item"):
        """
        variables: dict of variables
        elements: dict of elements
        functions: module of functions such as `math` or `numpy`

        get: 'item' if items in elements support el['key'], 'attr' if they support el.key
        """
        self.variables = variables
        self.functions = functions
        self.elements = elements
        grammar = calc_grammar
        if get == "attr":
            grammar = grammar.replace("getitem", "getattr")
        self.eval = Lark(grammar, parser="lalr", transformer=self).parse

    def assign_var(self, name, value):
        self.variables[name] = value
        return value

    def call(self, name, *args):
        ff = getattr(self.functions, name)
        return ff(*args)

    def getitem(self, name, key):
        return self.elements[name.value][key.value]

    def getattr(self, name, key):
        return getattr(self.elements[name], key)

    def var(self, name):
        try:
            return self.variables[name.value]
        except KeyError:
            raise Exception("Variable not found: %s" % name)


def test():
    import math
    from collections import defaultdict

    madx = MadxEval(defaultdict(lambda: 0), math, {})
    print(madx.eval("+1+2^-2"))
    print(madx.eval("a.b"))
    print(madx.eval("1+a.b*-3"))
    print(madx.eval("sin(3)^2"))


class Mix:
    __slots__ = ("_v", "_r")

    def __init__(self, data, refs):
        object.__setattr__(self, "_v", data)
        object.__setattr__(self, "_r", refs)

    def __getattr__(self, key):
        return self._v[key]

    def __getitem__(self, key):
        return self._v[key]

    def __setattr__(self, key, value):
        self._r[key] = value

    def __setitem__(self, key, value):
        self._r[key] = value

    def _eval(self, expr):
        return self._r._eval(expr)


class View:
    def __init__(self, obj, ref, evaluator):
        object.__setattr__(self, "_obj", obj)
        object.__setattr__(self, "_ref", ref)
        object.__setattr__(self, "_eval", evaluator)

    def _get_viewed_object(self):
        return object.__getattribute__(self, "_obj")

    @property
    def __class__(self):
        # return type("View",(self._obj.__class__,),{})
        return self._obj.__class__

    def get_value(self, key=None):
        if not hasattr(self._obj, "keys") and not hasattr(self._obj, "_xofields"):
            raise ValueError("get_value not supported for this object")

        if key is None:
            if hasattr(self._obj, "_xofields"):
                return {kk: self.get_value(kk) for kk in self._obj._xofields}
            else:
                return {kk: self.get_value(kk) for kk in dir(self._obj)}

        if hasattr(self._obj, "__iter__"):
            return self._obj[key]
        else:
            return getattr(self._obj, key)

    def get_expr(self, key=None, index=None):
        if index is not None:
            if key is None:
                raise ValueError('`key` must be provided when `index` is provided.')
            return getattr(self._ref, key)[index]._expr

        if not hasattr(self._obj, "keys") and not hasattr(self._obj, "_xofields"):
            # is an array
            return self._ref[key]._expr

        if key is None:
            if hasattr(self._obj, "_xofields"):
                return {kk: self.get_expr(kk) for kk in self._obj._xofields}
            else:
                return {kk: self.get_expr(kk) for kk in dir(self._obj)}

        if hasattr(self._obj, "__iter__"):
            return self._ref[key]._expr
        else:
            return getattr(self._ref, key)._expr


    def get_info(self, key=None):
        if not hasattr(self._obj, "keys") and not hasattr(self._obj, "_xofields"):
            raise ValueError("get_info not supported for this object")

        if key is None:
            print("Element of type: ", self._obj.__class__.__name__)
            self.get_table().show(header=False)
        else:
            if hasattr(self._obj, "__iter__"):
                return self._ref[key]._info()
            else:
                return getattr(self._ref, key)._info()

    def get_table(self):

        if not hasattr(self._obj, "keys") and not hasattr(self._obj, "_xofields"):
            raise ValueError("get_table not supported for this object")

        out_expr = self.get_expr()
        out_value = self.get_value()

        value = [out_value[kk] for kk in out_expr.keys()]
        for ii, vv in enumerate(value):
            if not(np.isscalar(vv)):
                value[ii] = str(vv)

        data = {
            "name": np.array(list(out_expr.keys()), dtype=object),
            "value": np.array(value, dtype=object),
            "expr": np.array(
                [str(out_expr[kk]) for kk in out_expr.keys()], dtype=object
            ),
        }
        return Table(data)

    def __getattr__(self, key):
        val = getattr(self._obj, key)
        if hasattr(val, "__setitem__"):
            return View(val, getattr(self._ref, key), self._eval)
        else:
            return val

    def __getitem__(self, key):
        val = self._obj[key]
        if hasattr(val, "__setitem__"):
            return View(val, self._ref[key], self._eval)
        else:
            return val

    def __setattr__(self, key, value):
        if isinstance(value, str):
            value = self._eval(value)
        setattr(self._ref, key, value)

    def __setitem__(self, key, value):
        if isinstance(value, str):
            value = self._eval(value)
        self._ref[key] = value

    def __repr__(self):
        return f"View of {self._obj!r}"

    def __dir__(self):
        return dir(self._obj)

    def __len__(self):
        return len(self._obj)

    def __add__(self, other):
        return self._obj + other

    def __radd__(self, other):
        return other + self._obj

    def __sub__(self, other):
        return self._obj - other

    def __rsub__(self, other):
        return other - self._obj

    def __mul__(self, other):
        return self._obj * other

    def __rmul__(self, other):
        return other * self._obj

    def __truediv__(self, other):
        return self._obj / other

    def __rtruediv__(self, other):
        return other / self._obj

    def __floordiv__(self, other):
        return self._obj // other

    def __rfloordiv__(self, other):
        return other // self._obj

    def __mod__(self, other):
        return self._obj % other

    def __rmod__(self, other):
        return other % self._obj

    def __pow__(self, other):
        return self._obj**other

    def __rpow__(self, other):
        return other**self._obj

    def __eq__(self, value: object) -> bool:
        return self._obj == value

    def __ne__(self, value: object) -> bool:
        return self._obj != value

    def __lt__(self, value: object) -> bool:
        return self._obj < value

    def __le__(self, value: object) -> bool:
        return self._obj <= value

    def __gt__(self, value: object) -> bool:
        return self._obj > value

    def __ge__(self, value: object) -> bool:
        return self._obj >= value


class MadxEnv:
    def __init__(self, mad=None):
        self._variables = defaultdict(lambda: 0)
        self._elements = {}
        self.manager = Manager()
        self._vref = self.manager.ref(self._variables, "v")
        self._eref = self.manager.ref(self._elements, "e")
        self._fref = self.manager.ref(math, "f")
        self.madexpr = MadxEval(self._vref, self._fref, self._eref).eval
        self.madeval = MadxEval(self._variables, math, self._elements).eval
        self.v = Mix(self._variables, self._vref)
        self.e = Mix(self._elements, self._eref)
        if mad is not None:
            self.read_state(mad)

    def dump(self):
        return {
            "variables": self._variables,
            "elements": self._elements,
            "xdeps": self.manager.dump(),
        }

    def load(self, data):
        self._variables.update(data["variables"])
        self._elements.update(data["elements"])
        self.manager.load(data["xdeps"])

    def to_json(self, filename):
        import json

        json.dump(self.dump(), open(filename, "w"))

    @classmethod
    def from_json(cls, filename):
        import json

        self = cls()
        self.load(json.load(open(filename)))
        return self

    def read_state(self, mad):
        elemdata = AttrDict()
        elem = mad.beam
        for parname, par in elem.cmdpar.items():
            elemdata[parname] = par.value
        self._elements["beam"] = elemdata

        for name, par in mad.globals.cmdpar.items():
            if par.expr is None:
                self._variables[name] = par.value
            else:
                self._vref[name] = self.madexpr(par.expr)

        for name, elem in mad.elements.items():
            elemdata = AttrDict()
            for parname, par in elem.cmdpar.items():
                elemdata[parname] = par.value
            self._elements[name] = elemdata

        for name, elem in mad.elements.items():
            for parname, par in elem.cmdpar.items():
                if par.expr is not None:
                    if par.dtype == 12:  # handle lists
                        for ii, ee in enumerate(par.expr):
                            if ee is not None:
                                self._eref[name][parname][ii] = self.madexpr(ee)
                    else:
                        self._eref[name][parname] = self.madexpr(par.expr)


def to_madx(expr):
    if expr.__class__.__name__ == "NegExpr":
        return f"(-{to_madx(expr._arg)})"
    elif expr.__class__.__name__ == "PosExpr":
        return f"{to_madx(expr._arg)}"
    elif expr.__class__.__name__ == "AddExpr":
        return f"({to_madx(expr._lhs)}+{to_madx(expr._rhs)})"
    elif expr.__class__.__name__ == "SubExpr":
        return f"({to_madx(expr._lhs)}-{to_madx(expr._rhs)})"
    elif expr.__class__.__name__ == "MulExpr":
        return f"({to_madx(expr._lhs)}*{to_madx(expr._rhs)})"
    elif expr.__class__.__name__ == "DivExpr":
        return f"({to_madx(expr._lhs)}/{to_madx(expr._rhs)})"
    elif expr.__class__.__name__ == "PowExpr":
        return f"({to_madx(expr._lhs)}^{to_madx(expr._rhs)})"
    elif expr.__class__.__name__ == "ItemRef":  # shortcut
        return f"{expr._key}"
    else:
        return repr(expr)


if __name__ == "__main__":
    test()
